"""
Module containing error views.
"""

import logging
import sys

from openapi_core.validation.schemas.exceptions import InvalidSchemaValue
from pyramid.httpexceptions import HTTPMethodNotAllowed
from pyramid.view import notfound_view_config
from pyramid.view import view_config
from pyramid_openapi3 import RequestValidationError
from pyramid_openapi3 import ResponseValidationError
from pyramid_openapi3 import extract_errors
from pyramid_openapi3 import openapi_validation_error
from skosprovider.exceptions import ProviderUnavailableException
from sqlalchemy.exc import IntegrityError

from atramhasis.errors import SkosRegistryNotFoundException
from atramhasis.errors import ValidationError
from atramhasis.protected_resources import ProtectedResourceException

log = logging.getLogger(__name__)


@notfound_view_config(renderer='json')
def failed_not_found(exc, request):
    """
    View invoked when a resource could not be found.
    """
    log.debug(exc.explanation)
    request.response.status_int = 404
    return {'message': exc.explanation}


@view_config(context=SkosRegistryNotFoundException, renderer='json')
def failed_skos(exc, request):
    """
    View invoked when Atramhasis can't find a SKOS registry.
    """
    log.error(exc.value, exc_info=sys.exc_info())
    request.response.status_int = 500
    return {'message': exc.value}


@view_config(context=ValidationError, renderer='json')
def failed_validation(exc, request):
    """
    View invoked when bad data was submitted to Atramhasis.
    """
    log.debug(f"'message': {exc.value}, 'errors': {exc.errors}")
    request.response.status_int = 400
    return {'message': exc.value, 'errors': exc.errors}


@view_config(context=ProtectedResourceException, renderer='json')
def protected(exc, request):
    """
    when a protected operation is called on a resource that is still referenced
    """
    log.warning(f"'message': {exc.value}, 'referenced_in': {exc.referenced_in}")
    request.response.status_int = 409
    return {'message': exc.value, 'referenced_in': exc.referenced_in}


@view_config(context=ProviderUnavailableException, renderer='json')
def provider_unavailable(exc, request):
    """
    View invoked when ProviderUnavailableException was raised.
    """
    log.error(exc, exc_info=sys.exc_info())
    request.response.status_int = 503
    return {'message': exc.message}


@view_config(context=IntegrityError, renderer='json')
def data_integrity(exc, request):
    """
    View invoked when IntegrityError was raised.
    """
    log.warning(exc)
    request.response.status_int = 409
    return {
        'message': 'this operation violates the data integrity and could not be executed'
    }


@view_config(context=Exception, renderer='json')
def failed(exc, request):  # pragma no cover
    """
    View invoked when bad data was submitted to Atramhasis.
    """
    log.error(exc, exc_info=sys.exc_info())
    request.response.status_int = 500
    return {'message': 'unexpected server error'}


@view_config(context=HTTPMethodNotAllowed, renderer='json')
def failed_not_method_not_allowed(exc, request):
    """
    View invoked when a method is not allowed.
    """
    log.debug(exc.explanation)
    request.response.status_int = 405
    return {'message': exc.explanation}


@view_config(context=RequestValidationError, renderer="json")
@view_config(context=ResponseValidationError, renderer="json")
def failed_openapi_validation(exc, request):
    try:
        errors = [
            _handle_validation_error(validation_error)
            for error in exc.errors
            if isinstance(error, InvalidSchemaValue)
            for validation_error in error.schema_errors
        ]
        # noinspection PyTypeChecker
        errors.extend(
            [
                f'{error.get("field")}: {error.get("message")}'
                for error in
                list(extract_errors(request, exc.errors))
            ]
        )
        request.response.status_int = 400
        return {"message": "Request was not valid for schema.", "errors": errors}
    except Exception:
        log.exception("Issue with exception handling.")
        return openapi_validation_error(exc, request)


def _handle_validation_error(error, path=""):
    if error.validator in ("anyOf", "oneOf", "allOf"):
        for schema_type in ("anyOf", "oneOf", "allOf"):
            if schema_type not in error.schema:
                continue
            schemas = error.schema.get(schema_type)
            break
        else:
            return None

        response = []
        for i, schema in enumerate(schemas):
            schema.pop("x-scope", None)
            errors = [
                sub_error
                for sub_error in error.context
                if sub_error.relative_schema_path[0] == i
            ]
            if error.path:
                schema = {".".join(str(p) for p in error.path): schema}
            response.append(
                {
                    "schema": schema,
                    "errors": [_handle_validation_error(error) for error in errors],
                }
            )
        return {schema_type: response}
    if path:
        path += "."
    path += ".".join(str(item) for item in error.path)
    if not path:
        path = "<root>"
    return f"{path}: {error.message}"



"""language_und

Revision ID: 184f1bbcb916
Revises: 6dfc3e2324aa
Create Date: 2017-07-25 16:38:39.439673

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.orm import Session
from sqlalchemy.sql import column
from sqlalchemy.sql import table

# revision identifiers, used by Alembic.
revision = "184f1bbcb916"
down_revision = "6dfc3e2324aa"


language_table = table("language", column("id", sa.String), column("name", sa.String))


def upgrade():

    op.bulk_insert(language_table, [{"id": "und", "name": "Undetermined"}])


def downgrade():
    connection = op.get_bind()
    session = Session(bind=connection)
    connection.execute(language_table.delete().where(language_table.c.id == "und"))
    session.flush()



MATERIAALDATA = [
    {
        "broader": [],
        "id": 1,
        "labels": [
            {"label": "aardewerk", "language": "nl-BE", "type": "prefLabel"},
            {"label": "ceramiek", "language": "nl-BE", "type": "altLabel"},
            {"label": "keramiek", "language": "nl-BE", "type": "altLabel"},
        ],
        "matches": {"broad": [], "close": [], "exact": [], "narrow": [], "related": []},
        "member_of": [],
        "narrower": [2, 5, 6],
        "notes": [
            {
                "language": "nl-BE",
                "markup": None,
                "note": (
                    "Aardewerk of keramiek is gebakken klei. Aardewerk wordt gebruikt"
                    " voor het vervaardigen van vaatwerk, bouwmaterialen, beeldjes,"
                    " enz\u2026"
                ),
                "type": "scopeNote",
            }
        ],
        "related": [],
        "sources": [
            {
                "citation": (
                    "Nederlandse Art &amp; Architecture Thesaurus, <a"
                    ' href="http://browser.aat-ned.nl/">http://browser.aat-ned.nl/</a>'
                ),
                "markup": "HTML",
            }
        ],
        "subordinate_arrays": [],
        "type": "concept",
        "uri": "https://id.erfgoed.net/thesauri/materialen/1",
    },
    {
        "broader": [8],
        "id": 48,
        "labels": [{"label": "aluminium", "language": "nl-BE", "type": "prefLabel"}],
        "matches": {"broad": [], "close": [], "exact": [], "narrow": [], "related": []},
        "member_of": [],
        "narrower": [],
        "notes": [
            {
                "language": "nl-BE",
                "markup": None,
                "note": (
                    "Aluminium is een metaal uit de derde groep van het periodiek"
                    " systeem (chemisch element Al). Het is dof zilverachtig van kleur"
                    " en zeer licht."
                ),
                "type": "scopeNote",
            }
        ],
        "related": [],
        "sources": [
            {
                "citation": (
                    "Van Dale groot woordenboek van de Nederlandse taal. 14de editie,"
                    " Utrecht/Antwerpen, 2005"
                ),
                "markup": None,
            }
        ],
        "subordinate_arrays": [],
        "type": "concept",
        "uri": "https://id.erfgoed.net/thesauri/materialen/48",
    },
    {
        "broader": [32],
        "id": 33,
        "labels": [
            {"label": "amber", "language": "nl-BE", "type": "prefLabel"},
            {"label": "barnsteen", "language": "nl-BE", "type": "altLabel"},
        ],
        "matches": {"broad": [], "close": [], "exact": [], "narrow": [], "related": []},
        "member_of": [],
        "narrower": [],
        "notes": [
            {
                "language": "nl-BE",
                "markup": None,
                "note": (
                    "Amber is een fossiele hars afkomstig uit naaldbomen en wordt"
                    " gebruikt voor het maken van kralen, hangers, knopen en dergelijke"
                    " meer."
                ),
                "type": "scopeNote",
            }
        ],
        "related": [],
        "sources": [{"citation": "VIOE", "markup": None}],
        "subordinate_arrays": [],
        "type": "concept",
        "uri": "https://id.erfgoed.net/thesauri/materialen/33",
    },
    {
        "broader": [],
        "id": 38,
        "labels": [{"label": "beton", "language": "nl-BE", "type": "prefLabel"}],
        "matches": {"broad": [], "close": [], "exact": [], "narrow": [], "related": []},
        "member_of": [],
        "narrower": [39],
        "notes": [
            {
                "language": "nl-BE",
                "markup": None,
                "note": (
                    "Beton is een kunstmatig materiaal dat lijkt op steen en als"
                    " bouwmateriaal gebruikt wordt. Het bestaat uit cement of kalk"
                    " waaraan zand, grind of steenslag wordt toegevoegd."
                ),
                "type": "scopeNote",
            }
        ],
        "related": [],
        "sources": [{"citation": "VIOE", "markup": None}],
        "subordinate_arrays": [],
        "type": "concept",
        "uri": "https://id.erfgoed.net/thesauri/materialen/38",
    },
    {
        "broader": [9],
        "id": 10,
        "labels": [{"label": "bladgoud", "language": "nl-BE", "type": "prefLabel"}],
        "matches": {"broad": [], "close": [], "exact": [], "narrow": [], "related": []},
        "member_of": [],
        "narrower": [],
        "notes": [
            {
                "language": "nl-BE",
                "markup": None,
                "note": (
                    "Bladgoud is een term die verwijst naar platen goud die tot een"
                    " zeer geringe dikte (meestal circa 0,1 micrometer) zijn gehamerd"
                    " of gewalst."
                ),
                "type": "scopeNote",
            }
        ],
        "related": [],
        "sources": [
            {
                "citation": (
                    "Nederlandse Art &amp; Architecture Thesaurus, <a"
                    ' href="http://browser.aat-ned.nl/">http://browser.aat-ned.nl/</a>'
                ),
                "markup": "HTML",
            }
        ],
        "subordinate_arrays": [],
        "type": "concept",
        "uri": "https://id.erfgoed.net/thesauri/materialen/10",
    },
    {
        "broader": [21],
        "id": 23,
        "labels": [{"label": "bont", "language": "nl-BE", "type": "prefLabel"}],
        "matches": {"broad": [], "close": [], "exact": [], "narrow": [], "related": []},
        "member_of": [],
        "narrower": [],
        "notes": [
            {
                "language": "nl-BE",
                "markup": None,
                "note": (
                    "Bont is een gelooide dierlijke huid, dicht bezet met haren. Het"
                    " wordt voornamelijk gebruikt voor het maken van kleding."
                ),
                "type": "scopeNote",
            }
        ],
        "related": [],
        "sources": [
            {
                "citation": (
                    "Nederlandse Art &amp; Architecture Thesaurus, <a"
                    ' href="http://browser.aat-ned.nl/">http://browser.aat-ned.nl/</a>'
                ),
                "markup": "HTML",
            }
        ],
        "subordinate_arrays": [],
        "type": "concept",
        "uri": "https://id.erfgoed.net/thesauri/materialen/23",
    },
    {
        "broader": [21],
        "id": 24,
        "labels": [{"label": "botmateriaal", "language": "nl-BE", "type": "prefLabel"}],
        "matches": {"broad": [], "close": [], "exact": [], "narrow": [], "related": []},
        "member_of": [],
        "narrower": [25, 26],
        "notes": [
            {
                "language": "nl-BE",
                "markup": None,
                "note": (
                    "Botmateriaal of been is een combinatie van organisch en"
                    " anorganisch materiaal die hoofdzakelijk bestaat uit collageen en"
                    " calciumfosfaat."
                ),
                "type": "scopeNote",
            }
        ],
        "related": [],
        "sources": [
            {
                "citation": (
                    "Nederlandse Art &amp; Architecture Thesaurus, <a"
                    ' href="http://browser.aat-ned.nl/">http://browser.aat-ned.nl/</a>'
                ),
                "markup": "HTML",
            }
        ],
        "subordinate_arrays": [],
        "type": "concept",
        "uri": "https://id.erfgoed.net/thesauri/materialen/24",
    },
    {
        "broader": [13],
        "id": 14,
        "labels": [{"label": "brons", "language": "nl-BE", "type": "prefLabel"}],
        "matches": {"broad": [], "close": [], "exact": [], "narrow": [], "related": []},
        "member_of": [],
        "narrower": [],
        "notes": [
            {
                "language": "nl-BE",
                "markup": None,
                "note": "Brons is een legering van koper en tin.",
                "type": "scopeNote",
            }
        ],
        "related": [],
        "sources": [],
        "subordinate_arrays": [],
        "type": "concept",
        "uri": "https://id.erfgoed.net/thesauri/materialen/14",
    },
    {
        "broader": [],
        "id": 49,
        "labels": [{"label": "cement", "language": "nl-BE", "type": "prefLabel"}],
        "matches": {"broad": [], "close": [], "exact": [], "narrow": [], "related": []},
        "member_of": [],
        "narrower": [],
        "notes": [
            {
                "language": "nl-BE",
                "markup": None,
                "note": (
                    "Cement is de benaming voor verschillende stoffen die een snel"
                    " verhardend bindmiddel voor bouwwerken (mortel) opleveren."
                ),
                "type": "scopeNote",
            }
        ],
        "related": [],
        "sources": [
            {
                "citation": (
                    "Van Dale groot woordenboek van de Nederlandse taal. 14de editie,"
                    " Utrecht/Antwerpen, 2005"
                ),
                "markup": None,
            }
        ],
        "subordinate_arrays": [],
        "type": "concept",
        "uri": "https://id.erfgoed.net/thesauri/materialen/49",
    },
    {
        "broader": [24],
        "id": 25,
        "labels": [
            {"label": "dierlijk botmateriaal", "language": "nl-BE", "type": "prefLabel"}
        ],
        "matches": {"broad": [], "close": [], "exact": [], "narrow": [], "related": []},
        "member_of": [],
        "narrower": [],
        "notes": [
            {
                "language": "nl-BE",
                "markup": None,
                "note": (
                    "Botmateriaal of been is een combinatie van organisch en"
                    " anorganisch materiaal die hoofdzakelijk bestaat uit collageen en"
                    " calciumfosfaat; het vormt het geraamte van de meeste gewervelde"
                    " dieren."
                ),
                "type": "scopeNote",
            }
        ],
        "related": [],
        "sources": [
            {
                "citation": (
                    "Nederlandse Art &amp; Architecture Thesaurus, <a"
                    ' href="http://browser.aat-ned.nl/">http://browser.aat-ned.nl/</a>'
                ),
                "markup": "HTML",
            }
        ],
        "subordinate_arrays": [],
        "type": "concept",
        "uri": "https://id.erfgoed.net/thesauri/materialen/25",
    },
    {
        "broader": [38],
        "id": 39,
        "labels": [
            {"label": "gewapend beton", "language": "nl-BE", "type": "prefLabel"}
        ],
        "matches": {"broad": [], "close": [], "exact": [], "narrow": [], "related": []},
        "member_of": [],
        "narrower": [],
        "notes": [
            {
                "language": "nl-BE",
                "markup": None,
                "note": "Gewapend beton is een combinatie van beton en staal.",
                "type": "scopeNote",
            }
        ],
        "related": [],
        "sources": [],
        "subordinate_arrays": [],
        "type": "concept",
        "uri": "https://id.erfgoed.net/thesauri/materialen/39",
    },
    {
        "broader": [21],
        "id": 27,
        "labels": [{"label": "gewei", "language": "nl-BE", "type": "prefLabel"}],
        "matches": {"broad": [], "close": [], "exact": [], "narrow": [], "related": []},
        "member_of": [],
        "narrower": [],
        "notes": [
            {
                "language": "nl-BE",
                "markup": None,
                "note": (
                    "Gewei wordt gevormd door enkele uitstulpingen van been op de"
                    " voorhoofdsbeenderen van gehoefde dieren die behoren tot de"
                    " hertenfamilie. Het wordt jaarlijks afgeworpen. Gewei wordt door"
                    " de mens gebruikt als grondstof voor het maken van artefacten."
                ),
                "type": "scopeNote",
            }
        ],
        "related": [],
        "sources": [
            {
                "citation": (
                    "Nederlandse Art &amp; Architecture Thesaurus, <a"
                    ' href="http://browser.aat-ned.nl/">http://browser.aat-ned.nl/</a>'
                ),
                "markup": "HTML",
            }
        ],
        "subordinate_arrays": [],
        "type": "concept",
        "uri": "https://id.erfgoed.net/thesauri/materialen/27",
    },
    {
        "broader": [],
        "id": 7,
        "labels": [{"label": "glas", "language": "nl-BE", "type": "prefLabel"}],
        "matches": {"broad": [], "close": [], "exact": [], "narrow": [], "related": []},
        "member_of": [],
        "narrower": [],
        "notes": [
            {
                "language": "nl-BE",
                "markup": None,
                "note": (
                    "Glas is een amorfe, anorganische substantie die wordt gemaakt door"
                    " silica (siliciumdioxide) te fuseren met een basisoxide, meestal"
                    " transparant maar vaak ook doorzichtig of ondoorschijnend."
                ),
                "type": "scopeNote",
            }
        ],
        "related": [],
        "sources": [
            {
                "citation": (
                    "Nederlandse Art &amp; Architecture Thesaurus, <a"
                    ' href="http://browser.aat-ned.nl/">http://browser.aat-ned.nl/</a>'
                ),
                "markup": "HTML",
            }
        ],
        "subordinate_arrays": [],
        "type": "concept",
        "uri": "https://id.erfgoed.net/thesauri/materialen/7",
    },
    {
        "broader": [8],
        "id": 9,
        "labels": [{"label": "goud", "language": "nl-BE", "type": "prefLabel"}],
        "matches": {"broad": [], "close": [], "exact": [], "narrow": [], "related": []},
        "member_of": [],
        "narrower": [10],
        "notes": [
            {
                "language": "nl-BE",
                "markup": None,
                "note": (
                    "Te gebruiken voor het zuivere metaalelement met het symbool Au."
                    " Het is een zacht, zwaar, chemisch inactief en geel metaal en"
                    " wordt al sinds de oudheid als edel beschouwd. Het dient in vele"
                    " culturen als de basis voor materi\xeble handelswaarden. Ook te"
                    " gebruiken voor het metaal wanneer het wordt bewerkt en vervormd"
                    " om, meestal in combinatie met andere stoffen, verschillende"
                    " voorwerpen en materialen te maken."
                ),
                "type": "scopeNote",
            }
        ],
        "related": [],
        "sources": [
            {
                "citation": (
                    "Nederlandse Art &amp; Architecture Thesaurus, <a"
                    ' href="http://browser.aat-ned.nl/">http://browser.aat-ned.nl/</a>'
                ),
                "markup": "HTML",
            }
        ],
        "subordinate_arrays": [],
        "type": "concept",
        "uri": "https://id.erfgoed.net/thesauri/materialen/9",
    },
    {
        "broader": [21],
        "id": 28,
        "labels": [{"label": "hoorn", "language": "nl-BE", "type": "prefLabel"}],
        "matches": {"broad": [], "close": [], "exact": [], "narrow": [], "related": []},
        "member_of": [],
        "narrower": [],
        "notes": [
            {
                "language": "nl-BE",
                "markup": None,
                "note": (
                    "Hoorn is een uitgroei van aangepaste huid of zeer compact haar,"
                    " vaak in paren die uitsteken vanaf de voorhoofdsbotten aan beide"
                    " zijden van de schedel en bestaat uit een permanente opperhuid van"
                    " keratine. De benen kern van een hoorn wordt hoornpit genoemd."
                ),
                "type": "scopeNote",
            },
            {
                "language": "nl-BE",
                "markup": None,
                "note": "Te onderscheiden van 'gewei', dat bestaat uit aangepast been.",
                "type": "note",
            },
        ],
        "related": [],
        "sources": [
            {
                "citation": (
                    "Nederlandse Art &amp; Architecture Thesaurus, <a"
                    ' href="http://browser.aat-ned.nl/">http://browser.aat-ned.nl/</a>'
                ),
                "markup": "HTML",
            }
        ],
        "subordinate_arrays": [],
        "type": "concept",
        "uri": "https://id.erfgoed.net/thesauri/materialen/28",
    },
    {
        "broader": [32],
        "id": 35,
        "labels": [{"label": "hout", "language": "nl-BE", "type": "prefLabel"}],
        "matches": {"broad": [], "close": [], "exact": [], "narrow": [], "related": []},
        "member_of": [],
        "narrower": [],
        "notes": [
            {
                "language": "nl-BE",
                "markup": None,
                "note": (
                    "Hout is het voornaamste bestanddeel van bomen en struiken. Hout is"
                    " een belangrijke grondstof en wordt voor allerlei toepassingen"
                    " gebruikt, zoals voor bouwconstructies, als brandstof, voor het"
                    " maken van papier, enz\u2026"
                ),
                "type": "scopeNote",
            }
        ],
        "related": [],
        "sources": [
            {
                "citation": (
                    "Nederlandse Art &amp; Architecture Thesaurus, <a"
                    ' href="http://browser.aat-ned.nl/">http://browser.aat-ned.nl/</a>'
                ),
                "markup": "HTML",
            }
        ],
        "subordinate_arrays": [],
        "type": "concept",
        "uri": "https://id.erfgoed.net/thesauri/materialen/35",
    },
    {
        "broader": [32],
        "id": 36,
        "labels": [{"label": "houtskool", "language": "nl-BE", "type": "prefLabel"}],
        "matches": {"broad": [], "close": [], "exact": [], "narrow": [], "related": []},
        "member_of": [],
        "narrower": [],
        "notes": [
            {
                "language": "nl-BE",
                "markup": None,
                "note": (
                    "Houtskool is een onzuivere vorm van grafietkoolstof die ontstaat"
                    " als reststof bij de gedeeltelijke verbranding van koolstofhoudend"
                    " materiaal, of bij verhitting van dit materiaal waarbij zuurstof"
                    " in beperkte hoeveelheid aanwezig is."
                ),
                "type": "scopeNote",
            }
        ],
        "related": [],
        "sources": [
            {
                "citation": (
                    "Nederlandse Art &amp; Architecture Thesaurus, <a"
                    ' href="http://browser.aat-ned.nl/">http://browser.aat-ned.nl/</a>'
                ),
                "markup": "HTML",
            }
        ],
        "subordinate_arrays": [],
        "type": "concept",
        "uri": "https://id.erfgoed.net/thesauri/materialen/36",
    },
    {
        "broader": [21],
        "id": 29,
        "labels": [{"label": "huid", "language": "nl-BE", "type": "prefLabel"}],
        "matches": {"broad": [], "close": [], "exact": [], "narrow": [], "related": []},
        "member_of": [],
        "narrower": [30],
        "notes": [],
        "related": [],
        "sources": [],
        "subordinate_arrays": [],
        "type": "concept",
        "uri": "https://id.erfgoed.net/thesauri/materialen/29",
    },
    {
        "broader": [8],
        "id": 11,
        "labels": [{"label": "ijzer", "language": "nl-BE", "type": "prefLabel"}],
        "matches": {"broad": [], "close": [], "exact": [], "narrow": [], "related": []},
        "member_of": [],
        "narrower": [],
        "notes": [
            {
                "language": "nl-BE",
                "markup": None,
                "note": (
                    "Te gebruiken voor het zuivere metaalelement met het symbool Fe."
                    " Het is een glimmend, zilverachtig zacht metaal dat roest wanneer"
                    " het aan vochtige lucht wordt blootgesteld. Gebruik ook voor dit"
                    " metaal als het wordt verwerkt en gevormd, meestal in combinatie"
                    " met andere stoffen, om diverse objecten en materialen te maken."
                ),
                "type": "scopeNote",
            }
        ],
        "related": [],
        "sources": [
            {
                "citation": (
                    "Nederlandse Art &amp; Architecture Thesaurus, <a"
                    ' href="http://browser.aat-ned.nl/">http://browser.aat-ned.nl/</a>'
                ),
                "markup": "HTML",
            }
        ],
        "subordinate_arrays": [],
        "type": "concept",
        "uri": "https://id.erfgoed.net/thesauri/materialen/11",
    },
    {
        "broader": [8],
        "id": 12,
        "labels": [{"label": "koper", "language": "nl-BE", "type": "prefLabel"}],
        "matches": {"broad": [], "close": [], "exact": [], "narrow": [], "related": []},
        "member_of": [],
        "narrower": [13],
        "notes": [
            {
                "language": "nl-BE",
                "markup": None,
                "note": (
                    "Te gebruiken voor het zuivere metaalelement met het symbool Cu."
                    " Het is roodachtig van kleur en is zeer smeedbaar en kneedbaar."
                    " Ook te gebruiken voor het metaal wanneer het wordt bewerkt en"
                    " gevormd om, meestal in combinatie met andere stoffen,"
                    " verschillende voorwerpen en materialen te maken."
                ),
                "type": "scopeNote",
            }
        ],
        "related": [],
        "sources": [
            {
                "citation": (
                    "Nederlandse Art &amp; Architecture Thesaurus, <a"
                    ' href="http://browser.aat-ned.nl/">http://browser.aat-ned.nl/</a>'
                ),
                "markup": "HTML",
            }
        ],
        "subordinate_arrays": [],
        "type": "concept",
        "uri": "https://id.erfgoed.net/thesauri/materialen/12",
    },
    {
        "broader": [12],
        "id": 13,
        "labels": [
            {"label": "koperlegeringen", "language": "nl-BE", "type": "prefLabel"}
        ],
        "matches": {"broad": [], "close": [], "exact": [], "narrow": [], "related": []},
        "member_of": [],
        "narrower": [14, 15],
        "notes": [
            {
                "language": "nl-BE",
                "markup": None,
                "note": (
                    "Een koperlegering is een vast mengsel van koper met andere"
                    " metalen."
                ),
                "type": "scopeNote",
            }
        ],
        "related": [],
        "sources": [{"citation": "VIOE", "markup": None}],
        "subordinate_arrays": [],
        "type": "concept",
        "uri": "https://id.erfgoed.net/thesauri/materialen/13",
    },
    {
        "broader": [],
        "id": 20,
        "labels": [{"label": "kunststof", "language": "nl-BE", "type": "prefLabel"}],
        "matches": {"broad": [], "close": [], "exact": [], "narrow": [], "related": []},
        "member_of": [],
        "narrower": [],
        "notes": [
            {
                "language": "nl-BE",
                "markup": None,
                "note": (
                    "Een kunststof is een langs chemische weg gemaakte stof die"
                    " gebruikt wordt voor de vervaardiging van de meest verschillende"
                    " zaken. In de scheepsbouw gaat het om een composiet van glasvezel"
                    " (of andere meer moderne vezels) met polyester of epoxy, al dan"
                    " niet in combinatie met een core-materiaal."
                ),
                "type": "scopeNote",
            }
        ],
        "related": [],
        "sources": [
            {
                "citation": (
                    "1) Van Dale groot woordenboek van de Nederlandse taal. 14de"
                    " editie, Utrecht/Antwerpen, 2005\n2) Onroerend Erfgoed"
                ),
                "markup": None,
            }
        ],
        "subordinate_arrays": [],
        "type": "concept",
        "uri": "https://id.erfgoed.net/thesauri/materialen/20",
    },
    {
        "broader": [43],
        "id": 44,
        "labels": [
            {"label": "kwartsiet van Tienen", "language": "nl-BE", "type": "prefLabel"}
        ],
        "matches": {"broad": [], "close": [], "exact": [], "narrow": [], "related": []},
        "member_of": [],
        "narrower": [],
        "notes": [
            {
                "language": "nl-BE",
                "markup": None,
                "note": (
                    "Het kwartsiet van Tienen is een metamorf gesteente waarbij de"
                    " oorspronkelijke sedimentaire kwartskorrels door secundaire"
                    " aangroei met elkaar zijn vergroeid. Dit verklaart de geringe"
                    " porositeit en sterke cohesie van dit gesteente."
                ),
                "type": "scopeNote",
            }
        ],
        "related": [],
        "sources": [
            {
                "citation": (
                    "Gullentops F. &amp; Wouters L., Delfstoffen in Vlaanderen,"
                    " Ministerie van de Vlaamse gemeenschap, Departement EWBL."
                ),
                "markup": None,
            }
        ],
        "subordinate_arrays": [],
        "type": "concept",
        "uri": "https://id.erfgoed.net/thesauri/materialen/44",
    },
    {
        "broader": [42],
        "id": 43,
        "labels": [
            {
                "label": "kwartsitisch lithisch materiaal",
                "language": "nl-BE",
                "type": "prefLabel",
            }
        ],
        "matches": {"broad": [], "close": [], "exact": [], "narrow": [], "related": []},
        "member_of": [],
        "narrower": [44, 45],
        "notes": [],
        "related": [],
        "sources": [],
        "subordinate_arrays": [],
        "type": "concept",
        "uri": "https://id.erfgoed.net/thesauri/materialen/43",
    },
    {
        "broader": [29],
        "id": 30,
        "labels": [{"label": "leer", "language": "nl-BE", "type": "prefLabel"}],
        "matches": {"broad": [], "close": [], "exact": [], "narrow": [], "related": []},
        "member_of": [],
        "narrower": [],
        "notes": [
            {
                "language": "nl-BE",
                "markup": None,
                "note": (
                    "Leer is de huid of het vel van een dier, die is gelooid om tegen"
                    " bederf te beschermen en in droge toestand relatief zacht en"
                    " soepel te maken."
                ),
                "type": "scopeNote",
            }
        ],
        "related": [],
        "sources": [
            {
                "citation": (
                    "Nederlandse Art &amp; Architecture Thesaurus, <a"
                    ' href="http://browser.aat-ned.nl/">http://browser.aat-ned.nl/</a>'
                ),
                "markup": "HTML",
            }
        ],
        "subordinate_arrays": [],
        "type": "concept",
        "uri": "https://id.erfgoed.net/thesauri/materialen/30",
    },
    {
        "broader": [40],
        "id": 42,
        "labels": [
            {"label": "lithisch materiaal", "language": "nl-BE", "type": "prefLabel"}
        ],
        "matches": {"broad": [], "close": [], "exact": [], "narrow": [], "related": []},
        "member_of": [],
        "narrower": [43, 46],
        "notes": [],
        "related": [],
        "sources": [],
        "subordinate_arrays": [],
        "type": "concept",
        "uri": "https://id.erfgoed.net/thesauri/materialen/42",
    },
    {
        "broader": [8],
        "id": 16,
        "labels": [{"label": "lood", "language": "nl-BE", "type": "prefLabel"}],
        "matches": {"broad": [], "close": [], "exact": [], "narrow": [], "related": []},
        "member_of": [],
        "narrower": [],
        "notes": [
            {
                "language": "nl-BE",
                "markup": None,
                "note": (
                    "Te gebruiken voor het zuivere metaalelement met het symbool Pb."
                    " Het metaal is zacht, kneedbaar en vaalgrijs van kleur. Ook te"
                    " gebruiken voor het metaal wanneer het wordt bewerkt en gevormd"
                    " om, meestal in combinatie met andere stoffen, verscheidene"
                    " voorwerpen en materialen te maken."
                ),
                "type": "scopeNote",
            }
        ],
        "related": [],
        "sources": [
            {
                "citation": (
                    "Nederlandse Art &amp; Architecture Thesaurus, <a"
                    ' href="http://browser.aat-ned.nl/">http://browser.aat-ned.nl/</a>'
                ),
                "markup": "HTML",
            }
        ],
        "subordinate_arrays": [],
        "type": "concept",
        "uri": "https://id.erfgoed.net/thesauri/materialen/16",
    },
    {
        "broader": [24],
        "id": 26,
        "labels": [
            {
                "label": "menselijk botmateriaal",
                "language": "nl-BE",
                "type": "prefLabel",
            }
        ],
        "matches": {"broad": [], "close": [], "exact": [], "narrow": [], "related": []},
        "member_of": [],
        "narrower": [],
        "notes": [
            {
                "language": "nl-BE",
                "markup": None,
                "note": (
                    "Botmateriaal of been is een combinatie van organisch en/of"
                    " anorganisch materiaal die hoofdzakelijk bestaat uit collageen en"
                    " calciumfosfaat; het vormt het geraamte van de meeste gewervelden,"
                    " waaronder de mens."
                ),
                "type": "scopeNote",
            }
        ],
        "related": [],
        "sources": [
            {
                "citation": (
                    "Nederlandse Art &amp; Architecture Thesaurus, <a"
                    ' href="http://browser.aat-ned.nl/">http://browser.aat-ned.nl/</a>'
                ),
                "markup": "HTML",
            }
        ],
        "subordinate_arrays": [],
        "type": "concept",
        "uri": "https://id.erfgoed.net/thesauri/materialen/26",
    },
    {
        "broader": [13],
        "id": 15,
        "labels": [{"label": "messing", "language": "nl-BE", "type": "prefLabel"}],
        "matches": {"broad": [], "close": [], "exact": [], "narrow": [], "related": []},
        "member_of": [],
        "narrower": [],
        "notes": [
            {
                "language": "nl-BE",
                "markup": None,
                "note": (
                    "Messing is een legering van koper en zink, meestal met koper als"
                    " het voornaamste element en met zink als maximaal 40% van het"
                    " gewicht."
                ),
                "type": "scopeNote",
            }
        ],
        "related": [],
        "sources": [
            {
                "citation": (
                    "Nederlandse Art &amp; Architecture Thesaurus, <a"
                    ' href="http://browser.aat-ned.nl/">http://browser.aat-ned.nl/</a>'
                ),
                "markup": "HTML",
            }
        ],
        "subordinate_arrays": [],
        "type": "concept",
        "uri": "https://id.erfgoed.net/thesauri/materialen/15",
    },
    {
        "broader": [],
        "id": 8,
        "labels": [{"label": "metaal", "language": "nl-BE", "type": "prefLabel"}],
        "matches": {"broad": [], "close": [], "exact": [], "narrow": [], "related": []},
        "member_of": [],
        "narrower": [9, 11, 12, 16, 17, 18, 19, 48],
        "notes": [
            {
                "language": "nl-BE",
                "markup": None,
                "note": (
                    "Metaal omvat een grote groep stoffen die meestal een kenmerkende"
                    " glans vertonen. Het zijn goede geleiders van elektriciteit en"
                    " warmte, ze zijn ondoorschijnend, kunnen smelten en zijn meestal"
                    " pletbaar of kneedbaar."
                ),
                "type": "scopeNote",
            }
        ],
        "related": [],
        "sources": [
            {
                "citation": (
                    "Nederlandse Art &amp; Architecture Thesaurus, <a"
                    ' href="http://browser.aat-ned.nl/">http://browser.aat-ned.nl/</a>'
                ),
                "markup": "HTML",
            }
        ],
        "subordinate_arrays": [],
        "type": "concept",
        "uri": "https://id.erfgoed.net/thesauri/materialen/8",
    },
    {
        "broader": [40],
        "id": 41,
        "labels": [{"label": "natuursteen", "language": "nl-BE", "type": "prefLabel"}],
        "matches": {"broad": [], "close": [], "exact": [], "narrow": [], "related": []},
        "member_of": [],
        "narrower": [],
        "notes": [
            {
                "language": "nl-BE",
                "markup": None,
                "note": (
                    "Natuursteen is een term die wordt gebruik voor gesteenten die in"
                    " de natuur worden aangetroffen en die, na eventuele bewerking,"
                    " kunnen dienen als bouwmateriaal. Hieronder vallen steensoorten"
                    " zoals graniet, tefriet, Doornikse kalksteen, zandsteen, Balegemse"
                    " kalkzandsteen, Gobertangesteen, arduin (blauwe hardsteen),"
                    " ijzerzandsteen, veldsteen, Ypresiaanse nummilietenkalksteen,"
                    " Avendersteen, schieffer, enz..."
                ),
                "type": "scopeNote",
            }
        ],
        "related": [],
        "sources": [
            {
                "citation": (
                    "1) Nederlandse Art &amp; Architecture Thesaurus, <a"
                    ' href="http://browser.aat-ned.nl/">http://browser.aat-ned.nl/</a>'
                    " 2) VIOE"
                ),
                "markup": "HTML",
            }
        ],
        "subordinate_arrays": [],
        "type": "concept",
        "uri": "https://id.erfgoed.net/thesauri/materialen/41",
    },
    {
        "broader": [],
        "id": 21,
        "labels": [
            {"label": "organisch materiaal", "language": "nl-BE", "type": "prefLabel"}
        ],
        "matches": {"broad": [], "close": [], "exact": [], "narrow": [], "related": []},
        "member_of": [],
        "narrower": [23, 24, 27, 28, 29, 32],
        "notes": [
            {
                "language": "nl-BE",
                "markup": None,
                "note": (
                    "Organisch materiaal is materiaal dat koolstof bevat, inclusief de"
                    " materialen die afkomstig zijn van levende organismen."
                ),
                "type": "scopeNote",
            }
        ],
        "related": [],
        "sources": [
            {
                "citation": (
                    "Nederlandse Art &amp; Architecture Thesaurus, <a"
                    ' href="http://browser.aat-ned.nl/">http://browser.aat-ned.nl/</a>'
                ),
                "markup": "HTML",
            }
        ],
        "subordinate_arrays": [],
        "type": "concept",
        "uri": "https://id.erfgoed.net/thesauri/materialen/21",
    },
    {
        "broader": [1],
        "id": 2,
        "labels": [{"label": "pijpaarde", "language": "nl-BE", "type": "prefLabel"}],
        "matches": {"broad": [], "close": [], "exact": [], "narrow": [], "related": []},
        "member_of": [],
        "narrower": [],
        "notes": [
            {
                "language": "nl-BE",
                "markup": None,
                "note": (
                    "Pijpaarde is een bepaalde soort fijne, witbakkende klei die"
                    " gebruikt werd voor het maken van pijpen, beeldjes, fijn aardewerk"
                    " enz..."
                ),
                "type": "scopeNote",
            }
        ],
        "related": [],
        "sources": [{"citation": "VIOE", "markup": None}],
        "subordinate_arrays": [],
        "type": "concept",
        "uri": "https://id.erfgoed.net/thesauri/materialen/2",
    },
    {
        "broader": [21],
        "id": 32,
        "labels": [
            {"label": "plantaardig materiaal", "language": "nl-BE", "type": "prefLabel"}
        ],
        "matches": {"broad": [], "close": [], "exact": [], "narrow": [], "related": []},
        "member_of": [],
        "narrower": [33, 35, 36, 37],
        "notes": [],
        "related": [],
        "sources": [],
        "subordinate_arrays": [],
        "type": "concept",
        "uri": "https://id.erfgoed.net/thesauri/materialen/32",
    },
    {
        "broader": [],
        "id": 50,
        "labels": [{"label": "pleister", "language": "nl-BE", "type": "prefLabel"}],
        "matches": {
            "broad": [],
            "close": [],
            "exact": ["http://vocab.getty.edu/aat/300014922"],
            "narrow": [],
            "related": [],
        },
        "member_of": [],
        "narrower": [],
        "notes": [
            {
                "language": "nl-BE",
                "markup": None,
                "note": (
                    "Pleister verwijst naar zacht, plastisch materiaal dat kan worden"
                    " uitgerold of uitgestreken op een muur, plafond of ander oppervlak"
                    " waarop het vervolgens verhardt. In de context van kunst en"
                    " architectuur verwijst het meestal specifiek naar een mengsel van"
                    " water, kalk en zand, vaak in combinatie met andere materialen"
                    " zoals dierenhaar, waardoor het materiaal meer kracht, textuur en"
                    " - als het oppervlak daarna wordt geschilderd - poreusheid krijgt."
                ),
                "type": "scopeNote",
            }
        ],
        "related": [],
        "sources": [{"citation": "Art &amp; Architecture Thesaurus", "markup": None}],
        "subordinate_arrays": [],
        "type": "concept",
        "uri": "https://id.erfgoed.net/thesauri/materialen/50",
    },
    {
        "broader": [1],
        "id": 5,
        "labels": [{"label": "porselein", "language": "nl-BE", "type": "prefLabel"}],
        "matches": {"broad": [], "close": [], "exact": [], "narrow": [], "related": []},
        "member_of": [],
        "narrower": [],
        "notes": [
            {
                "language": "nl-BE",
                "markup": None,
                "note": (
                    "Porselein verwijst naar een bepaald soort hardgebakken aardewerk"
                    " van vuurbestendige witte klei, kaolien, waaraan veldspaat en"
                    " kwarts worden toegevoegd."
                ),
                "type": "scopeNote",
            }
        ],
        "related": [],
        "sources": [{"citation": "VIOE", "markup": None}],
        "subordinate_arrays": [],
        "type": "concept",
        "uri": "https://id.erfgoed.net/thesauri/materialen/5",
    },
    {
        "broader": [],
        "id": 31,
        "labels": [{"label": "schelp", "language": "nl-BE", "type": "prefLabel"}],
        "matches": {"broad": [], "close": [], "exact": [], "narrow": [], "related": []},
        "member_of": [],
        "narrower": [],
        "notes": [
            {
                "language": "nl-BE",
                "markup": None,
                "note": (
                    "De het uitwendig skelet van een weekdier samengesteld uit kalk of"
                    " andere mineralen."
                ),
                "type": "scopeNote",
            }
        ],
        "related": [],
        "sources": [],
        "subordinate_arrays": [],
        "type": "concept",
        "uri": "https://id.erfgoed.net/thesauri/materialen/31",
    },
    {
        "broader": [42],
        "id": 46,
        "labels": [
            {"label": "silex", "language": "nl-BE", "type": "prefLabel"},
            {"label": "vuursteen", "language": "nl-BE", "type": "altLabel"},
        ],
        "matches": {"broad": [], "close": [], "exact": [], "narrow": [], "related": []},
        "member_of": [],
        "narrower": [],
        "notes": [
            {
                "language": "nl-BE",
                "markup": None,
                "note": (
                    "Een fijnkorrelig, compact sedimentair gesteente dat bestaat uit"
                    " kwartskristallen."
                ),
                "type": "scopeNote",
            }
        ],
        "related": [],
        "sources": [
            {
                "citation": (
                    "Nederlandse Art &amp; Architecture Thesaurus, <a"
                    ' href="http://browser.aat-ned.nl/">http://browser.aat-ned.nl/</a>'
                ),
                "markup": "HTML",
            }
        ],
        "subordinate_arrays": [],
        "type": "concept",
        "uri": "https://id.erfgoed.net/thesauri/materialen/46",
    },
    {
        "broader": [8],
        "id": 17,
        "labels": [{"label": "staal", "language": "nl-BE", "type": "prefLabel"}],
        "matches": {"broad": [], "close": [], "exact": [], "narrow": [], "related": []},
        "member_of": [],
        "narrower": [],
        "notes": [
            {
                "language": "nl-BE",
                "markup": None,
                "note": "Staal is een legering van ijzer en koolstof.",
                "type": "scopeNote",
            }
        ],
        "related": [],
        "sources": [],
        "subordinate_arrays": [],
        "type": "concept",
        "uri": "https://id.erfgoed.net/thesauri/materialen/17",
    },
    {
        "broader": [],
        "id": 40,
        "labels": [{"label": "steen", "language": "nl-BE", "type": "prefLabel"}],
        "matches": {"broad": [], "close": [], "exact": [], "narrow": [], "related": []},
        "member_of": [],
        "narrower": [41, 42],
        "notes": [
            {
                "language": "nl-BE",
                "markup": None,
                "note": "Steen is een harde stof met een minerale samenstelling.",
                "type": "scopeNote",
            }
        ],
        "related": [],
        "sources": [],
        "subordinate_arrays": [],
        "type": "concept",
        "uri": "https://id.erfgoed.net/thesauri/materialen/40",
    },
    {
        "broader": [1],
        "id": 6,
        "labels": [{"label": "steengoed", "language": "nl-BE", "type": "prefLabel"}],
        "matches": {"broad": [], "close": [], "exact": [], "narrow": [], "related": []},
        "member_of": [],
        "narrower": [],
        "notes": [
            {
                "language": "nl-BE",
                "markup": None,
                "note": (
                    "Steengoed is een soort aardewerk dat gemaakt is van klei en"
                    " smeltbaar gesteente. Het materiaal wordt zolang gebakken tot er"
                    " gedeeltelijke verglazing optreedt, waardoor het materiaal"
                    " waterdicht wordt."
                ),
                "type": "scopeNote",
            }
        ],
        "related": [],
        "sources": [
            {
                "citation": (
                    "Nederlandse Art &amp; Architecture Thesaurus, <a"
                    ' href="http://browser.aat-ned.nl/">http://browser.aat-ned.nl/</a>'
                ),
                "markup": "HTML",
            }
        ],
        "subordinate_arrays": [],
        "type": "concept",
        "uri": "https://id.erfgoed.net/thesauri/materialen/6",
    },
    {
        "broader": [32],
        "id": 37,
        "labels": [{"label": "textiel", "language": "nl-BE", "type": "prefLabel"}],
        "matches": {"broad": [], "close": [], "exact": [], "narrow": [], "related": []},
        "member_of": [],
        "narrower": [],
        "notes": [
            {
                "language": "nl-BE",
                "markup": None,
                "note": (
                    "Textiel is een materiaal dat tot stand is gekomen door middel van"
                    " weven, vervilten, knopen, twijnen of andersoortige bewerking van"
                    " natuurlijke of synthetische vezels zodanig dat zij bijeenblijven."
                ),
                "type": "scopeNote",
            }
        ],
        "related": [],
        "sources": [
            {
                "citation": (
                    "Nederlandse Art &amp; Architecture Thesaurus, <a"
                    ' href="http://browser.aat-ned.nl/">http://browser.aat-ned.nl/</a>'
                ),
                "markup": "HTML",
            }
        ],
        "subordinate_arrays": [],
        "type": "concept",
        "uri": "https://id.erfgoed.net/thesauri/materialen/37",
    },
    {
        "broader": [8],
        "id": 18,
        "labels": [{"label": "tin", "language": "nl-BE", "type": "prefLabel"}],
        "matches": {"broad": [], "close": [], "exact": [], "narrow": [], "related": []},
        "member_of": [],
        "narrower": [],
        "notes": [
            {
                "language": "nl-BE",
                "markup": None,
                "note": (
                    "Te gebruiken voor het zuivere metaalelement met het symbool Sn."
                    " Het zilverachtig witte metaal is zacht en buigzaam. Ook te"
                    " gebruiken voor dit metaal wanneer het wordt bewerkt en gevormd"
                    " om, meestal in combinatie met andere stoffen, diverse voorwerpen"
                    " en materialen te maken."
                ),
                "type": "scopeNote",
            }
        ],
        "related": [],
        "sources": [
            {
                "citation": (
                    "Nederlandse Art &amp; Architecture Thesaurus, <a"
                    ' href="http://browser.aat-ned.nl/">http://browser.aat-ned.nl/</a>'
                ),
                "markup": "HTML",
            }
        ],
        "subordinate_arrays": [],
        "type": "concept",
        "uri": "https://id.erfgoed.net/thesauri/materialen/18",
    },
    {
        "broader": [43],
        "id": 45,
        "labels": [
            {"label": "wommersomkwartsiet", "language": "nl-BE", "type": "prefLabel"}
        ],
        "matches": {"broad": [], "close": [], "exact": [], "narrow": [], "related": []},
        "member_of": [],
        "narrower": [],
        "notes": [
            {
                "language": "nl-BE",
                "markup": None,
                "note": (
                    "Wommersomkwartsiet vertoont een textuur van hoekige kwartskorrels"
                    " in een fijnkorrelig kwartscement ingebed. Deze gesteentesoort"
                    " komt schijnbaar alleen voor in Wommersom, nabij Tienen. Het"
                    " vertoont enige gelijkenis met silex en werd alsdusdanig ook"
                    " gebruikt om prehistorische werktuigen te maken."
                ),
                "type": "scopeNote",
            }
        ],
        "related": [],
        "sources": [
            {
                "citation": (
                    "Gullentops F. &amp; Wouters L., Delfstoffen in Vlaanderen,"
                    " Ministerie van de Vlaamse gemeenschap, Departement EWBL."
                ),
                "markup": None,
            }
        ],
        "subordinate_arrays": [],
        "type": "concept",
        "uri": "https://id.erfgoed.net/thesauri/materialen/45",
    },
    {
        "broader": [8],
        "id": 19,
        "labels": [{"label": "zilver", "language": "nl-BE", "type": "prefLabel"}],
        "matches": {"broad": [], "close": [], "exact": [], "narrow": [], "related": []},
        "member_of": [],
        "narrower": [],
        "notes": [
            {
                "language": "nl-BE",
                "markup": None,
                "note": (
                    "Te gebruiken voor het zuivere metaalelement met het symbool Ag."
                    " Het metaal is smeedbaar, vervormbaar en wit van kleur met een"
                    " kenmerkende glans en wordt als edel beschouwd. Ook te gebruiken"
                    " voor dit metaal wanneer het wordt bewerkt en gevormd om, meestal"
                    " in combinatie met andere stoffen, diverse voorwerpen en"
                    " materialen te maken."
                ),
                "type": "scopeNote",
            }
        ],
        "related": [],
        "sources": [
            {
                "citation": (
                    "Nederlandse Art &amp; Architecture Thesaurus, <a"
                    ' href="http://browser.aat-ned.nl/">http://browser.aat-ned.nl/</a>'
                ),
                "markup": "HTML",
            }
        ],
        "subordinate_arrays": [],
        "type": "concept",
        "uri": "https://id.erfgoed.net/thesauri/materialen/19",
    },
]



from sqlalchemy import select

import tests
from atramhasis.data.models import ExpandStrategy
from atramhasis.data.models import IDGenerationStrategy
from atramhasis.data.models import Provider
from atramhasis.scripts import initializedb as script


def setUpModule():
    tests.setup_db(guarantee_empty=True)


class TestMigrateTests(tests.DbTest):
    def test_initialize_providers(self):
        script.initialize_providers(self.session)

        expected_ids = [
            'TREES',
            'GEOGRAPHY',
            'STYLE',
            'MATERIAL',
            'EVENTTYPE',
            'HERITAGETYPE',
            'PERIOD',
            'SPECIES',
            'BLUEBIRDS',
        ]
        expected_concept_scheme_ids = [1, 2, 3, 4, 5, 6, 7, 8, 9]
        for db_provider in self.session.execute(select(Provider)).scalars():  # type: Provider
            self.assertIn(db_provider.conceptscheme.id, expected_concept_scheme_ids)
            self.assertEqual(db_provider.id_generation_strategy, IDGenerationStrategy.NUMERIC)
            self.assertEqual(db_provider.expand_strategy, ExpandStrategy.RECURSE)
            self.assertIn(db_provider.id, expected_ids)



"""Add markup to source.

Revision ID: 3bcf11802900
Revises: 1a4d62b02630
Create Date: 2016-05-25 14:31:51.579412

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "3bcf11802900"
down_revision = "1a4d62b02630"


def upgrade():
    op.add_column("source", sa.Column("markup", sa.String(20)))


def downgrade():
    with op.batch_alter_table("source") as batch_op:
        batch_op.drop_column("markup")



"""init SKOS data

Revision ID: 4f1ee5c9c08
Revises: 3924a7ad2f8
Create Date: 2014-04-15 16:12:09.144433

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.sql import column
from sqlalchemy.sql import table

# revision identifiers, used by Alembic.
revision = "4f1ee5c9c08"
down_revision = "3924a7ad2f8"


notetype_table = table(
    "notetype", column("name", sa.String), column("description", sa.String)
)

labeltype_table = table(
    "labeltype", column("name", sa.String), column("description", sa.String)
)

language_table = table("language", column("id", sa.String), column("name", sa.String))


def upgrade():
    op.bulk_insert(
        notetype_table,
        [
            {"name": "changeNote", "description": "A change note."},
            {"name": "definition", "description": "A definition."},
            {"name": "editorialNote", "description": "An editorial note."},
            {"name": "example", "description": "An example."},
            {"name": "historyNote", "description": "A historynote."},
            {"name": "scopeNote", "description": "A scopenote."},
            {"name": "note", "description": "A note."},
        ],
    )

    op.bulk_insert(
        labeltype_table,
        [
            {"name": "hiddenLabel", "description": "A hidden label."},
            {"name": "altLabel", "description": "An alternative label."},
            {"name": "prefLabel", "description": "A preferred label."},
        ],
    )

    op.bulk_insert(
        language_table,
        [
            {"id": "la", "name": "Latin"},
            {"id": "nl", "name": "Dutch"},
            {"id": "vls", "name": "Flemish"},
            {"id": "en", "name": "English"},
            {"id": "fr", "name": "French"},
            {"id": "de", "name": "German"},
        ],
    )
    pass


def downgrade():
    op.execute(notetype_table.delete())
    op.execute(labeltype_table.delete())
    op.execute(language_table.delete())
    pass



import unittest
from unittest.mock import Mock

import colander
from pyramid import testing
from skosprovider_sqlalchemy.models import Collection
from skosprovider_sqlalchemy.models import Concept
from skosprovider_sqlalchemy.models import LabelType
from sqlalchemy.exc import NoResultFound

from atramhasis import validators
from atramhasis.errors import ValidationError
from atramhasis.data.models import IDGenerationStrategy
from atramhasis.validators import Concept as ConceptSchema
from atramhasis.validators import ConceptScheme as ConceptSchemeSchema
from atramhasis.validators import LanguageTag
from atramhasis.validators import concept_schema_validator
from atramhasis.validators import conceptscheme_schema_validator
from atramhasis.validators import languagetag_validator


class DummySkosManager:
    def get_thing(self, concept_id, conceptscheme_id):
        concept = Concept(concept_id=concept_id, conceptscheme_id=conceptscheme_id)
        if concept_id == "2":
            broader_concept = Concept(concept_id="1", conceptscheme_id=conceptscheme_id)
            concept.broader_concepts = {broader_concept}
            return concept
        if concept_id == "7":
            narrower_concept = Concept(concept_id="14", conceptscheme_id=conceptscheme_id)
            concept.narrower_concepts = {narrower_concept}
            return concept
        if concept_id == "666":
            concept = Collection(concept_id=concept_id, conceptscheme_id=conceptscheme_id)
            return concept
        if concept_id == "667":
            concept = Collection(concept_id=concept_id, conceptscheme_id=conceptscheme_id)
            memberof = Collection(concept_id="60", conceptscheme_id=conceptscheme_id)
            memberofs = set()
            memberofs.add(memberof)
            concept.member_of = memberofs
            return concept
        if concept_id == "62":
            concept = Collection(concept_id=concept_id, conceptscheme_id=conceptscheme_id)
            member = Collection(concept_id="666", conceptscheme_id=conceptscheme_id)
            members = set()
            members.add(member)
            concept.members = members
            return concept
        if concept_id == "777":
            if conceptscheme_id != "3":
                raise NoResultFound()
        return concept

    def get_all_label_types(self):
        return [
            LabelType('hiddenLabel', 'A hidden label.'),
            LabelType('altLabel', 'An alternative label.'),
            LabelType('prefLabel', 'A preferred label.'),
        ]


class DummyLanguagesManager:
    def count_languages(self, language_id):
        if language_id in ['af', 'flup']:
            return 0
        else:
            return 1

    def save(self, language):
        return language


class TestValidation(unittest.TestCase):
    def setUp(self):
        self.config = testing.setUp()
        self.request = testing.DummyRequest()
        self.request.data_managers = {
            'skos_manager': DummySkosManager(),
            'languages_manager': DummyLanguagesManager()
        }
        self.provider = Mock(conceptscheme_id=1, metadata={})
        self.concept_schema = ConceptSchema(
            validator=concept_schema_validator
        ).bind(
            request=self.request,
            provider=self.provider,
            validate_id_generation=True,
        )
        self.language = LanguageTag(
            validator=languagetag_validator
        ).bind(
            request=self.request,
            new=True
        )
        self.conceptscheme_schema = ConceptSchemeSchema(
            validator=conceptscheme_schema_validator
        ).bind(
            request=self.request
        )
        self.json_concept = {
            "narrower": [{"id": "8"}, {"id": "7"}, {"id": "9"}],
            "label": "Belgium",
            "type": "concept",
            "id": "4",
            "broader": [{"id": "2"}],
            "related": [{"id": "5"}],
            "labels": [{
                "label": "Belgium",
                "type": "prefLabel",
                "language": "en"
            }],
            "notes": [{
                "note": "een notitie",
                "type": "note",
                "language": "nl"
            }],
            "sources": [{
                "citation": "Van Daele K. 2014"
            }],
            "member_of": [{"id": "666"}]
        }
        self.json_collection = {
            "id": "0",
            "labels": [{
                "language": "nl-BE",
                "label": "Stijlen en culturen",
                "type": "prefLabel"
            }],
            "type": "collection",
            "label": "Stijlen en culturen",
            "members": [{"id": "61"}, {"id": "60"}],
            "notes": [{
                "note": "een notitie",
                "type": "note",
                "language": "nl"
            }],
            "member_of": [{"id": "666"}]
        }
        self.json_conceptscheme = {
            "labels": [{
                "language": "nl-BE",
                "label": "Stijlen en culturen",
                "type": "prefLabel"
            }],
            "label": "Stijlen en culturen",
            "notes": [{
                "note": "een notitie",
                "type": "note",
                "language": "nl"
            }],
            "sources": [{
                "citation": "Van Daele K. 2014"
            }]
        }

    def tearDown(self):
        testing.tearDown()

    def test_validation_conceptscheme(self):
        validated_conceptscheme = self.conceptscheme_schema.deserialize(self.json_conceptscheme)
        self.assertIsNotNone(validated_conceptscheme)
        self.assertEqual(1, len(validated_conceptscheme['labels']))
        self.assertEqual(1, len(validated_conceptscheme['notes']))
        self.assertEqual(1, len(validated_conceptscheme['sources']))

    def test_invalid_conceptscheme(self):
        self.json_conceptscheme.pop('labels')
        self.assertRaises(
            ValidationError, self.conceptscheme_schema.deserialize, self.json_conceptscheme
        )

    def test_validation_concept(self):
        validated_concept = self.concept_schema.deserialize(self.json_concept)
        self.assertIsNotNone(validated_concept)
        self.assertEqual(1, len(validated_concept['labels']))
        self.assertEqual(1, len(validated_concept['notes']))
        self.assertEqual(1, len(validated_concept['sources']))
        self.assertEqual(3, len(validated_concept['narrower']))
        self.assertEqual(1, len(validated_concept['broader']))
        self.assertEqual(1, len(validated_concept['related']))
        self.assertEqual(1, len(validated_concept['member_of']))

    def test_max_preflabels_2_en(self):
        self.json_concept['labels'].append({
            "label": "B",
            "type": "prefLabel",
            "language": "en"
        })
        with self.assertRaises(ValidationError) as e:
            self.concept_schema.deserialize(self.json_concept)

        self.assertIn(
            {'labels': 'Only one prefLabel per language allowed.'}, e.exception.errors
        )

    def test_max_preflabels_1_en_1_nl(self):
        self.json_concept['labels'].append({
            "label": "B",
            "type": "prefLabel",
            "language": "nl"
        })
        validated_concept = self.concept_schema.deserialize(self.json_concept)
        self.assertIsNotNone(validated_concept)

    def test_related_concept_type_collection(self):
        self.json_concept['related'].append({"id": 666})
        with self.assertRaises(ValidationError) as e:
            self.concept_schema.deserialize(self.json_concept)
        self.assertIn(
            {
                'related': 'A narrower, broader or related concept should '
                           'always be a concept, not a collection'
            },
            e.exception.errors
        )

    def test_collection_with_related(self):
        # Collections can not have related relations
        self.json_collection['related'] = []
        self.json_collection['related'].append({"id": 2})
        with self.assertRaises(ValidationError) as e:
            self.concept_schema.deserialize(self.json_collection)
        self.assertIn(
            {'related': 'Only concepts can have narrower/broader/related relations'},
            e.exception.errors
        )

    def test_narrower_concept_type_collection(self):
        self.json_concept['narrower'].append({"id": 666})
        with self.assertRaises(ValidationError) as e:
            self.concept_schema.deserialize(self.json_concept)
        self.assertIn(
            {
                'narrower': 'A narrower, broader or related concept should '
                            'always be a concept, not a collection'
            },
            e.exception.errors
        )
    
    def test_infer_concept_relations(self):
        self.json_concept['infer_concept_relations'] = True
        with self.assertRaises(ValidationError) as e:
            self.concept_schema.deserialize(self.json_concept)
        self.assertIn(
            {
                'infer_concept_relations': "'infer_concept_relations' can only "
                                           "be set for collections."
            },
            e.exception.errors
        )
        self.json_collection['infer_concept_relations'] = True
        self.concept_schema.deserialize(self.json_collection)

    def test_collection_with_narrower(self):
        # Collections can not have narrower relations
        self.json_collection['narrower'] = []
        self.json_collection['narrower'].append({"id": 2})
        with self.assertRaises(ValidationError) as e:
            self.concept_schema.deserialize(self.json_collection)
        self.assertIn(
            {'narrower': 'Only concepts can have narrower/broader/related relations'},
            e.exception.errors
        )

    def test_broader_concept_type_collection(self):
        self.json_concept['broader'].append({"id": 666})
        with self.assertRaises(ValidationError) as e:
            self.concept_schema.deserialize(self.json_concept)
        self.assertIn(
            {
                'broader': 'A narrower, broader or related concept should '
                           'always be a concept, not a collection'
            },
            e.exception.errors
        )

    def test_collection_with_broader(self):
        # Collections can not have broader relations
        self.json_collection['broader'] = []
        self.json_collection['broader'].append({"id": 2})
        with self.assertRaises(ValidationError):
            self.concept_schema.deserialize(self.json_collection)

    def test_related_concept_different_conceptscheme(self):
        self.json_concept['related'].append({"id": 777})
        with self.assertRaises(ValidationError) as e:
            self.concept_schema.deserialize(self.json_concept)
        self.assertIn(
            {
                'related': 'Concept not found, check concept_id. Please be aware '
                           'members should be within one scheme'
            },
            e.exception.errors
        )

    def test_narrower_concept_different_conceptscheme(self):
        self.json_concept['narrower'].append({"id": 777})
        with self.assertRaises(ValidationError) as e:
            self.concept_schema.deserialize(self.json_concept)
        self.assertIn(
            {
                'narrower': 'Concept not found, check concept_id. Please be aware '
                            'members should be within one scheme'
            },
            e.exception.errors
        )

    def test_narrower_concept_to_self(self):
        self.json_concept['narrower'].append({"id": 4})
        with self.assertRaises(ValidationError) as e:
            self.concept_schema.deserialize(self.json_concept)
        self.assertIn(
            {'narrower': 'A concept or collection cannot be related to itself'},
            e.exception.errors
        )

    def test_broader_concept_different_conceptscheme(self):
        self.json_concept['broader'].append({"id": 777})
        with self.assertRaises(ValidationError) as e:
            self.concept_schema.deserialize(self.json_concept)
        self.assertIn(
            {
                'broader': 'Concept not found, check concept_id. Please be aware '
                           'members should be within one scheme'
             },
            e.exception.errors
        )

    def test_broader_concept_hierarchy(self):
        self.json_concept['broader'].append({"id": 14})
        with self.assertRaises(ValidationError) as e:
            self.concept_schema.deserialize(self.json_concept)
        self.assertIn(
            {
                'broader': 'The broader concept of a concept must not itself '
                           'be a narrower concept of the concept being edited.'
            },
            e.exception.errors
        )

    def test_broader_concept_hierarchy_no_narrower(self):
        self.json_concept['broader'].append({"id": 8})
        self.json_concept['narrower'] = []
        validated_concept = self.concept_schema.deserialize(self.json_concept)
        self.assertIsNotNone(validated_concept)

    def test_narrower_concept_hierarchy(self):
        self.json_concept['narrower'].append({"id": 1})
        with self.assertRaises(ValidationError) as e:
            self.concept_schema.deserialize(self.json_concept)
        self.assertIn(
            {
                'narrower': 'The narrower concept of a concept must not itself '
                            'be a broader concept of the concept being edited.'
            },
            e.exception.errors
        )

    def test_narrower_concept_hierarchy_no_broader(self):
        self.json_concept['narrower'].append({"id": 1})
        self.json_concept['broader'] = []
        validated_concept = self.concept_schema.deserialize(self.json_concept)
        self.assertIsNotNone(validated_concept)

    def test_validation_collection(self):
        validated_collection = self.concept_schema.deserialize(self.json_collection)
        self.assertIsNotNone(validated_collection)
        self.assertEqual(2, len(validated_collection['members']))
        self.assertEqual(1, len(validated_collection['labels']))
        self.assertEqual(1, len(validated_collection['notes']))

    def test_member_concept_different_conceptscheme(self):
        self.json_collection['members'].append({"id": 777})
        with self.assertRaises(ValidationError) as e:
            self.concept_schema.deserialize(self.json_collection)
        self.assertIn(
            {
                'members': 'Concept not found, check concept_id. Please be aware '
                           'members should be within one scheme'
            },
            e.exception.errors
        )

    def test_label_type(self):
        self.json_concept['labels'].append({
            "label": "Belgium",
            "type": "altLabel",
            "language": "en"
        })
        validated_concept = self.concept_schema.deserialize(self.json_concept)
        self.assertIsNotNone(validated_concept)

    def test_label_type_invalid(self):
        self.json_concept['labels'].append({
            "label": "Belgium",
            "type": "testLabelInvalid",
            "language": "en"
        })
        with self.assertRaises(ValidationError) as e:
            self.concept_schema.deserialize(self.json_concept)
        self.assertIn({'labels': 'Invalid labeltype.'}, e.exception.errors)

    def test_label_language_invalid(self):
        self.json_concept['labels'].append({
            "label": "Belgium",
            "type": "altLabel",
            "language": "eng"
        })
        with self.assertRaises(ValidationError) as e:
            self.concept_schema.deserialize(self.json_concept)
        self.assertIn(
            {
                'labels': "Invalid language tag: Unknown code 'eng', "
                          "Missing language tag in 'eng'."
            },
            e.exception.errors
        )

    def test_label_language_missing(self):
        self.json_concept['labels'].append({
            "label": "Belgium",
            "type": "altLabel",
            "language": "af"
        })
        validated_concept = self.concept_schema.deserialize(self.json_concept)
        self.assertIsNotNone(validated_concept)

    def test_label_invalid(self):
        self.json_concept['labels'].append({
            "note": "Belgium",
            "type": "altLabel",
            "language": "en"
        })
        with self.assertRaises(colander.Invalid):
            self.concept_schema.deserialize(self.json_concept)

    def test_note_invalid(self):
        self.json_concept['notes'].append({
            "label": "een notitie",
            "type": 5,
            "language": "nl"
        })
        with self.assertRaises(colander.Invalid):
            self.concept_schema.deserialize(self.json_concept)

    def test_memberof_concept_type_collection(self):
        # A Collection/Concept can be a member_of a Collection
        validated_concept = self.concept_schema.deserialize(self.json_concept)
        self.assertIsNotNone(validated_concept)

    def test_memberof_concept_type_concept(self):
        # Nothing can be a member_of a Concept
        self.json_concept['member_of'].append({"id": 2})
        with self.assertRaises(ValidationError) as e:
            self.concept_schema.deserialize(self.json_concept)
        self.assertIn(
            {'member_of': 'A member_of parent should always be a collection'},
            e.exception.errors
        )

    def test_members_collection_unique(self):
        # A Collection is a Set (every element of the Collection should be unique).
        self.json_collection['members'].append({"id": 61})
        with self.assertRaises(ValidationError) as e:
            self.concept_schema.deserialize(self.json_collection)
        self.assertIn(
            {'members': 'All members of a collection should be unique.'},
            e.exception.errors
        )

    def test_concept_members(self):
        # A Concept does not have members.
        self.json_concept['members'] = []
        self.json_concept['members'].append({"id": 2})
        with self.assertRaises(ValidationError) as e:
            self.concept_schema.deserialize(self.json_concept)
        self.assertIn(
            {'members': 'Only collections can have members.'}, e.exception.errors
        )

    def test_memberof_concept_hierarchy_simple(self):
        # The hierarchy should not contain loops
        self.json_collection['members'].append({"id": 666})
        with self.assertRaises(ValidationError) as e:
            self.concept_schema.deserialize(self.json_collection)
        self.assertIn(
            {
                'member_of': 'The parent member_of collection of a concept must not '
                             'itself be a member of the concept being edited.'
            },
            e.exception.errors
        )

    def test_memberof_concept_hierarchy_deep(self):
        # The hierarchy should not contain loops
        self.json_collection['members'].append({"id": 62})
        with self.assertRaises(ValidationError) as e:
            self.concept_schema.deserialize(self.json_collection)
        self.assertIn(
            {
                'member_of': 'The parent member_of collection of a concept must not '
                             'itself be a member of the concept being edited.'
            },
            e.exception.errors
        )

    def test_members_concept_hierarchy_simple(self):
        # The hierarchy should not contain loops
        self.json_collection['member_of'].append({"id": 61})
        with self.assertRaises(ValidationError) as e:
            self.concept_schema.deserialize(self.json_collection)
        self.assertIn(
            {
                'members': 'The item of a members collection must not itself be a '
                           'parent of the concept/collection being edited.'
            },
            e.exception.errors
        )

    def test_members_concept_hierarchy_deep(self):
        # The hierarchy should not contain loops
        self.json_collection['member_of'].append({"id": 667})
        with self.assertRaises(ValidationError) as e:
            self.concept_schema.deserialize(self.json_collection)
        self.assertIn(
            {
                'members': 'The item of a members collection must not itself be a '
                           'parent of the concept/collection being edited.'
            },
            e.exception.errors
        )

    def test_min_labels_rule_empty_labels(self):
        self.json_concept['labels'] = []
        with self.assertRaises(ValidationError) as e:
            self.concept_schema.deserialize(self.json_concept)
        self.assertIn({'labels': 'At least one label is necessary'}, e.exception.errors)

    def test_min_labels_rule_no_labels(self):
        json_concept = {
            "narrower": [{"id": 8}, {"id": 7}, {"id": 9}],
            "type": "concept",
            "id": 4,
            "broader": [{"id": 2}],
            "related": [{"id": 5}],
            "notes": [{
                "note": "een notitie",
                "type": "note",
                "language": "nl"
            }],
            "member_of": [{"id": 666}]
        }
        with self.assertRaises(ValidationError) as e:
            self.concept_schema.deserialize(json_concept)
        self.assertIn({'labels': 'At least one label is necessary'}, e.exception.errors)

    def test_concept_matches_rule(self):
        json_concept = {
            "type": "collection",
            "labels": [{
                "language": "nl",
                "label": "Stijlen en culturen",
                "type": "prefLabel"
            }],
            "id": 4,
            "members": [{"id": 666}],
            "matches": {"exactMatch": ["urn:sample:666"], "broadMatch": ["urn:somewhere:93"]}
        }
        with self.assertRaises(ValidationError) as e:
            self.concept_schema.deserialize(json_concept)
        self.assertIn({'matches': 'Only concepts can have matches'}, e.exception.errors)

    def test_concept_matches_unique_rule(self):
        json_concept = {
            "type": "concept",
            "labels": [{
                "language": "nl",
                "label": "Stijlen en culturen",
                "type": "prefLabel"
            }],
            "id": 4,
            "member_of": [{"id": 666}],
            "matches": {"exact": ["urn:sample:666"], "broad": ["urn:sample:666"]}
        }
        with self.assertRaises(ValidationError) as e:
            self.concept_schema.deserialize(json_concept)

        self.assertIn(
            {'matches': 'All matches of a concept should be unique.'}, e.exception.errors
        )

    def test_concept_matches_unique_rule_pass(self):
        json_concept = {
            "type": "concept",
            "labels": [{
                "language": "nl",
                "label": "Stijlen en culturen",
                "type": "prefLabel"
            }],
            "id": 4,
            "member_of": [{"id": 666}],
            "matches": {"exactMatch": ["urn:sample:666"], "broadMatch": ["urn:sample:93"]}
        }
        validated_concept = self.concept_schema.deserialize(json_concept)
        self.assertIsNotNone(validated_concept)

    def test_languages_pass(self):
        json_language = {
            "id": "af",
            "name": "Afrikaans"
        }
        validated_language = self.language.deserialize(json_language)
        self.assertIsNotNone(validated_language)

    def test_languages_duplicate(self):
        json_language = {
            "id": "en",
            "name": "English"
        }
        with self.assertRaises(ValidationError) as e:
            self.language.deserialize(json_language)
        self.assertIn({'id': 'Duplicate language tag: en'}, e.exception.errors)

    def test_languages_edit_not_raise_duplicate(self):
        json_language = {
            "id": "en",
            "name": "English"
        }
        language = LanguageTag(
            validator=languagetag_validator
        ).bind(
            request=self.request,
            new=False
        )
        validated_language = language.deserialize(json_language)
        self.assertIsNotNone(validated_language)

    def test_languages_invalid(self):
        json_language = {
            "id": "flup",
            "name": "test"
        }
        with self.assertRaises(ValidationError) as e:
            self.language.deserialize(json_language)
        self.assertIn(
            {
                "id": "Invalid language tag: Unknown code 'flup', Missing "
                      "language tag in 'flup'."
            },
            e.exception.errors
        )

    def test_subordinate_arrays(self):
        self.json_concept['subordinate_arrays'] = [{"id": 667}]
        validated_json = self.concept_schema.deserialize(self.json_concept)
        self.assertIsNotNone(validated_json)

    def test_subordinate_arrays_no_concept(self):
        self.json_collection['subordinate_arrays'] = [{"id": 666}]
        with self.assertRaises(ValidationError) as e:
            self.concept_schema.deserialize(self.json_collection)
        self.assertIn(
            {'subordinate_arrays': 'Only concept can have subordinate arrays.'},
            e.exception.errors
        )

    def test_subordinate_arrays_no_collection(self):
        self.json_concept['subordinate_arrays'] = [{"id": 7}]
        with self.assertRaises(ValidationError) as e:
            self.concept_schema.deserialize(self.json_concept)
        self.assertIn(
            {'subordinate_arrays': 'A subordinate array should always be a collection'},
            e.exception.errors
        )

    def test_subordinate_arrays_hierarchy(self):
        self.json_concept['subordinate_arrays'] = [{"id": 666}]
        with self.assertRaises(ValidationError) as e:
            self.concept_schema.deserialize(self.json_concept)
        self.assertIn(
            {
                'subordinate_arrays': 'The subordinate_array collection of a concept must '
                                      'not itself be a parent of the concept being edited.'
            },
            e.exception.errors
        )

    def test_superordinates(self):
        self.json_collection['superordinates'] = [{"id": 7}]
        validated_json = self.concept_schema.deserialize(self.json_collection)
        self.assertIsNotNone(validated_json)

    def test_superordinates_no_concept(self):
        self.json_collection['superordinates'] = [{"id": 666}]
        with self.assertRaises(ValidationError) as e:
            self.concept_schema.deserialize(self.json_collection)
        self.assertIn(
            {'superordinates': 'A superordinate should always be a concept'},
            e.exception.errors
        )

    def test_superordinates_no_collection(self):
        self.json_concept['superordinates'] = [{"id": 7}]
        with self.assertRaises(ValidationError) as e:
            self.concept_schema.deserialize(self.json_concept)
        self.assertIn(
            {'superordinates': 'Only collection can have superordinates.'},
            e.exception.errors
        )

    def test_superordinates_hierarchy(self):
        self.json_collection['superordinates'] = [{"id": 61}]
        with self.assertRaises(ValidationError) as e:
            self.concept_schema.deserialize(self.json_collection)
        self.assertIn(
            {
                'superordinates': 'The superordinates of a collection must not itself '
                                  'be a member of the collection being edited.'
            },
            e.exception.errors
        )

    def test_html_in_notes(self):
        json_concept = {
            "narrower": [{"id": 8}, {"id": 7}, {"id": 9}],
            "label": "Belgium",
            "type": "concept",
            "id": 4,
            "broader": [{"id": 2}],
            "related": [{"id": 5}],
            "labels": [{
                "label": "Belgium",
                "type": "prefLabel",
                "language": "en"
            }],
            "notes": [{
                "note": "een <strong><h2>notitie</h2></strong>",
                "type": "note",
                "language": "nl"
            }],
            "sources": [{
                "citation": "Van Daele K. 2014"
            }],
            "member_of": [{"id": 666}]
        }
        validated_json = self.concept_schema.deserialize(json_concept)
        note = validated_json['notes'][0]
        self.assertEqual("een <strong>\nnotitie</strong>", note['note'])

    def test_html_in_sources(self):
        json_concept = {
            "narrower": [{"id": 8}, {"id": 7}, {"id": 9}],
            "label": "Belgium",
            "type": "concept",
            "id": 4,
            "broader": [{"id": 2}],
            "related": [{"id": 5}],
            "labels": [{
                "label": "Belgium",
                "type": "prefLabel",
                "language": "en"
            }],
            "notes": [{
                "note": "een notitie",
                "type": "note",
                "language": "nl"
            }],
            "sources": [{
                "citation": "Van Daele K. <strong><h2>2014</h2></strong>"
            }],
            "member_of": [{"id": 666}]
        }
        validated_json = self.concept_schema.deserialize(json_concept)
        source = validated_json['sources'][0]
        self.assertEqual("Van Daele K. <strong>\n2014</strong>", source['citation'])

    def test_id_generation_manual_no_id(self):
        concept_schema = self.concept_schema.bind(
            request=self.request,
            provider=self.provider,
            validate_id_generation=True,
        )
        self.provider.metadata["atramhasis.id_generation_strategy"] = (
            IDGenerationStrategy.MANUAL
        )
        del self.json_concept["id"]
        with self.assertRaises(ValidationError) as e:
            concept_schema.deserialize(self.json_concept)
        self.assertIn({'id': 'Required for this provider.'}, e.exception.errors)
        concept_schema = self.concept_schema.bind(
            request=self.request,
            provider=self.provider,
            validate_id_generation=False,
        )
        concept_schema.deserialize(self.json_concept)

    def test_id_generation_manual_not_unique(self):
        concept_schema = self.concept_schema.bind(
            request=self.request,
            provider=self.provider,
            validate_id_generation=True,
        )
        self.provider.metadata["atramhasis.id_generation_strategy"] = (
            IDGenerationStrategy.MANUAL
        )
        self.json_concept["id"] = "1"
        with self.assertRaises(ValidationError) as e:
            concept_schema.deserialize(self.json_concept)
        self.assertIn({'id': '1 already exists.'}, e.exception.errors)
        concept_schema = self.concept_schema.bind(
            request=self.request,
            provider=self.provider,
            validate_id_generation=False,
        )
        concept_schema.deserialize(self.json_concept)

    def test_validate_provider(self):
        """
        This looks like it does not validate much, but openapi validates the majority.

        We only need to manually validate what openapi cannot.
        """
        json_data = {'metadata': {}}
        validators.validate_provider_json(json_data)

    def test_validate_provider_forbidden_keys(self):
        forbidden_metadata_keys = (
            'default_language',
            'subject',
            'force_display_language',
            'atramhasis.force_display_language',
            'id_generation_strategy',
            'atramhasis.id_generation_strategy',
        )
        for bad_key in forbidden_metadata_keys:
            json_data = {'metadata': {bad_key: 'value'}}
            with self.assertRaises(ValidationError):
                validators.validate_provider_json(json_data)

    def test_validate_provider_bad_language(self):
        json_data = {'default_language': 'notalanguage'}
        with self.assertRaises(ValidationError):
            validators.validate_provider_json(json_data)
        json_data = {'force_display_language': 'notalanguage'}
        with self.assertRaises(ValidationError):
            validators.validate_provider_json(json_data)

    def test_validate_provider_update_with_wrong_id(self):
        json_data = {'id': 'notanid'}
        with self.assertRaises(ValidationError):
            validators.validate_provider_json(json_data, "test")



"""Add provider table

Revision ID: b2a7d7614973
Revises: 0978347c16f9
Create Date: 2023-03-22 15:10:27.781804

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "b2a7d7614973"
down_revision = "0978347c16f9"


def upgrade():
    op.create_table(
        "provider",
        sa.Column("id", sa.String(), nullable=False),
        sa.Column("conceptscheme_id", sa.Integer(), nullable=False),
        sa.Column("uri_pattern", sa.Text(), nullable=False),
        sa.Column("metadata", sa.JSON(), nullable=False),
        sa.Column(
            "expand_strategy",
            sa.Enum("RECURSE", "VISIT", name="expandstrategy"),
            nullable=True,
        ),
        sa.ForeignKeyConstraint(
            ["conceptscheme_id"],
            ["conceptscheme.id"],
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("id"),
    )


def downgrade():
    op.drop_table("provider")



"""add skos matches

Revision ID: 441c5a16ef8
Revises: 4f1ee5c9c08
Create Date: 2014-10-24 11:04:10.310845

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.sql import column
from sqlalchemy.sql import table

# revision identifiers, used by Alembic.
revision = "441c5a16ef8"
down_revision = "4f1ee5c9c08"


def upgrade():
    op.create_table(
        "matchtype",
        sa.Column("name", sa.String(length=20), nullable=False, primary_key=True),
        sa.Column("description", sa.Text(), nullable=True),
    )
    op.create_table(
        "match",
        sa.Column("uri", sa.String(length=512), nullable=False),
        sa.Column("concept_id", sa.Integer, nullable=False),
        sa.Column("matchtype_id", sa.String(length=20), nullable=False),
        sa.ForeignKeyConstraint(
            ["concept_id"],
            ["concept.id"],
        ),
        sa.ForeignKeyConstraint(
            ["matchtype_id"],
            ["matchtype.name"],
        ),
        sa.PrimaryKeyConstraint("uri", "concept_id", "matchtype_id"),
    )
    op.bulk_insert(
        table(
            "matchtype",
            column("name", sa.String),
            column("description", sa.Text),
        ),
        [
            {
                "name": "closeMatch",
                "description": "Indicates that two concepts are sufficiently similar that they can be used interchangeably in some information retrieval applications.",
            },
            {
                "name": "exactMatch",
                "description": "Indicates that there is a high degree of confidence that two concepts can be used interchangeably across a wide range of information retrieval applications.",
            },
            {
                "name": "broadMatch",
                "description": "Indicates that one concept has a broader match with another one.",
            },
            {
                "name": "narrowMatch",
                "description": "Indicates that one concept has a narrower match with another one.",
            },
            {
                "name": "relatedMatch",
                "description": "Indicates that there is an associative mapping between two concepts.",
            },
        ],
    )


def downgrade():
    op.drop_table("match")
    op.drop_table("matchtype")



"""
Testdata cotaining species

.. versionadded:: 0.4.4
"""


from skosprovider.providers import DictionaryProvider
from skosprovider.uri import UriPatternGenerator
from .species_data import SPECIESDATA

species = DictionaryProvider(
    {'id': 'SPECIES'},
    SPECIESDATA,
    uri_generator=UriPatternGenerator('https://id.erfgoed.net/thesauri/soorten/%s')
)



"""
Module containing errors generated by Atramhasis.
"""

from pyramid.httpexceptions import HTTPNotFound


class SkosRegistryNotFoundException(Exception):
    """
    Atramhasis could not find a SKOS registry.
    """
    def __init__(self, value='No SKOS registry found, please check your application setup'):
        self.value = value

    def __str__(self):
        return repr(self.value)


class ConceptSchemeNotFoundException(HTTPNotFound):
    """
    A ConceptScheme could not be found.
    """
    def __init__(self, scheme_id):
        self.value = 'No conceptscheme found with the given id ' + scheme_id

    def __str__(self):
        return repr(self.value)


class LanguageNotFoundException(HTTPNotFound):
    """
    A Language could not be found.
    """
    def __init__(self, scheme_id):
        self.value = 'No language found with the given id ' + scheme_id

    def __str__(self):
        return repr(self.value)


class ConceptNotFoundException(HTTPNotFound):
    """
    A Concept or Collection could not be found.
    """
    def __init__(self, c_id):
        self.value = 'No concept found with the given id ' + c_id

    def __str__(self):
        return repr(self.value)


class ValidationError(Exception):
    """
    Some data that was validated is invalid.
    """
    def __init__(self, value, errors):
        self.value = value
        self.errors = errors

    def __str__(self):
        return repr(self.value)


class DbNotFoundException(Exception):
    """
    Atramhasis could not find a database.
    """

    def __init__(self, value='No database found, please check your application setup'):
        self.value = value

    def __str__(self):
        return repr(self.value)


class SQLAlchemyProviderNotFoundException(HTTPNotFound):
    """
    SQLAlchemyProvider could not be found in the database.
    """

    def __init__(self, p_id):
        self.value = 'No SQLAlchemyProvider found with the given id ' + p_id

    def __str__(self):
        return repr(self.value)



from dogpile.cache import make_region

tree_region = make_region()
list_region = make_region()


def includeme(config):
    settings = config.registry.settings
    if not tree_region.is_configured:
        tree_region.configure_from_config(settings, 'cache.tree.')
    if not list_region.is_configured:
        list_region.configure_from_config(settings, 'cache.list.')


def invalidate_scheme_cache(conceptscheme_id):
    # TODO implement propper conceptscheme cache invalidation if this is ever possible
    tree_region.invalidate()


def invalidate_cache():
    tree_region.invalidate()



"""
Testdata cotaining materials: concrete, glass, metal, silex, ...

.. versionadded:: 0.2.0
"""


from skosprovider.providers import DictionaryProvider
from skosprovider.uri import UriPatternGenerator
from .materiaal_data import MATERIAALDATA

materials = DictionaryProvider(
    {'id': 'MATERIAL'},
    MATERIAALDATA,
    uri_generator=UriPatternGenerator('https://id.erfgoed.net/thesauri/materialen/%s')
)



"""add_sortlabel

Revision ID: 6dfc3e2324aa
Revises: b04fd493106b
Create Date: 2017-07-04 13:53:00.064535

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.orm import Session

# revision identifiers, used by Alembic.
revision = "6dfc3e2324aa"
down_revision = "b04fd493106b"


labeltype_table = sa.Table(
    "labeltype",
    sa.MetaData(),
    sa.Column("name", sa.String),
    sa.Column("description", sa.String),
)


def upgrade():

    op.bulk_insert(
        labeltype_table, [{"name": "sortLabel", "description": "A sorting label."}]
    )


def downgrade():
    connection = op.get_bind()
    session = Session(bind=connection)
    connection.execute(
        labeltype_table.delete().where(labeltype_table.c.name == "sortLabel")
    )
    session.flush()



import os

from paste.deploy import appconfig
from skosprovider_sqlalchemy.models import Collection
from skosprovider_sqlalchemy.models import Concept
from skosprovider_sqlalchemy.models import ConceptScheme
from skosprovider_sqlalchemy.models import Label
from skosprovider_sqlalchemy.models import LabelType
from skosprovider_sqlalchemy.models import Language
from skosprovider_sqlalchemy.models import Note
from skosprovider_sqlalchemy.models import NoteType
from skosprovider_sqlalchemy.models import Source
from skosprovider_sqlalchemy.models import Visitation
from sqlalchemy import select

from atramhasis.scripts import delete_scheme
from tests import DbTest
from tests import db_session
from tests import fill_db
from tests import setup_db

TEST_DIR = os.path.dirname(__file__)
settings = appconfig('config:' + os.path.join(TEST_DIR, '../conf_test.ini'))


def setUpModule():
    setup_db()
    fill_db()


class DeleteSchemeTest(DbTest):

    def test_delete(self):
        with db_session() as session:
            for scheme_id in range(1, 10):
                delete_scheme.delete_scheme(session, scheme_id)
            assert len(session.execute(select(ConceptScheme)).all()) == 0
            assert len(session.execute(select(Concept)).all()) == 0
            assert len(session.execute(select(Collection)).all()) == 0
            assert len(session.execute(select(Note)).all()) == 0
            assert len(session.execute(select(Source)).all()) == 0
            assert len(session.execute(select(Visitation)).all()) == 0
            assert len(session.execute(select(Label)).all()) == 0
            assert len(session.execute(select(LabelType)).all()) != 0
            assert len(session.execute(select(NoteType)).all()) != 0
            assert len(session.execute(select(Language)).all()) != 0



"""
Module containing utility functions used by Atramhasis.
"""
import contextlib
import copy
from collections import deque

from pyramid.httpexceptions import HTTPMethodNotAllowed
from skosprovider.skos import Collection
from skosprovider.skos import Concept
from skosprovider.skos import ConceptScheme
from skosprovider.skos import Label
from skosprovider.skos import Note
from skosprovider.skos import Source
from skosprovider.uri import UriPatternGenerator
from skosprovider_sqlalchemy.providers import SQLAlchemyProvider
from sqlalchemy import engine_from_config
from sqlalchemy import orm
from sqlalchemy.orm import sessionmaker

from atramhasis.data.models import Provider


def from_thing(thing):
    """
        Map a :class:`skosprovider_sqlalchemy.models.Thing` to a
        :class:`skosprovider.skos.Concept` or
        a :class:`skosprovider.skos.Collection`, depending on the type.

        :param skosprovider_sqlalchemy.models.Thing thing: Thing to map.
        :rtype: :class:`~skosprovider.skos.Concept` or
            :class:`~skosprovider.skos.Collection`.
        """
    if thing.type and thing.type == 'collection':
        return Collection(
            id=thing.concept_id,
            uri=thing.uri,
            concept_scheme=ConceptScheme(thing.conceptscheme.uri),
            labels=[
                Label(label.label, label.labeltype_id, label.language_id)
                for label in thing.labels
            ],
            notes=[
                Note(n.note, n.notetype_id, n.language_id)
                for n in thing.notes
            ],
            sources=[
                Source(s.citation)
                for s in thing.sources
            ],
            members=[member.concept_id for member in thing.members] if hasattr(thing, 'members') else [],
            member_of=[c.concept_id for c in thing.member_of],
            superordinates=[broader_concept.concept_id for broader_concept in thing.broader_concepts],
            infer_concept_relations=thing.infer_concept_relations
        )
    else:
        matches = {}
        for m in thing.matches:
            key = m.matchtype.name[:m.matchtype.name.find('Match')]
            if key not in matches:
                matches[key] = []
            matches[key].append(m.uri)
        return Concept(
            id=thing.concept_id,
            uri=thing.uri,
            concept_scheme=ConceptScheme(thing.conceptscheme.uri),
            labels=[
                Label(label.label, label.labeltype_id, label.language_id)
                for label in thing.labels
            ],
            notes=[
                Note(n.note, n.notetype_id, n.language_id)
                for n in thing.notes
            ],
            sources=[
                Source(s.citation)
                for s in thing.sources
            ],
            broader=[c.concept_id for c in thing.broader_concepts],
            narrower=[c.concept_id for c in thing.narrower_concepts],
            related=[c.concept_id for c in thing.related_concepts],
            member_of=[c.concept_id for c in thing.member_of],
            subordinate_arrays=[narrower_collection.concept_id for narrower_collection in thing.narrower_collections],
            matches=matches,
        )


def internal_providers_only(fn):
    """
    aspect oriented way to check if provider is internal when calling the decorated function

    :param fn: the decorated function
    :return: around advice
    :raises pyramid.httpexceptions.HTTPMethodNotAllowed: when provider is not internal
    """
    def advice(parent_object, *args, **kw):

        if isinstance(parent_object.provider, SQLAlchemyProvider) and \
                'external' not in parent_object.provider.get_metadata()['subject']:
            return fn(parent_object, *args, **kw)
        else:
            raise HTTPMethodNotAllowed()

    return advice


def update_last_visited_concepts(request, concept_data):
    deque_last_visited = deque(maxlen=4)
    deque_last_visited.extend(request.session.get('last_visited', []))
    try:
        # Try to remove concept from the queue to prevent double entries
        deque_last_visited.remove(concept_data)
    except ValueError:
        # Concept is not in the queue
        pass
    # Add concept to the queue
    deque_last_visited.append(concept_data)
    request.session['last_visited'] = list(deque_last_visited)


def label_sort(concepts, language='any'):
    if not concepts:
        return []
    return sorted(
        concepts, key=lambda concept: concept._sortkey(key='sortlabel',
                                                       language=language)
    )


def db_provider_to_skosprovider(db_provider: Provider) -> SQLAlchemyProvider:
    """Create a SQLAlchemyProvider from a atramhasis.data.models.Provider.

    :param db_provider: The Provider to use as basis for the SQLAlchemyProvider.
    :return: An SQLAlchemyProvider with the data from the `db_provider`
    """
    metadata = copy.deepcopy(db_provider.meta)
    metadata["conceptscheme_id"] = db_provider.conceptscheme_id
    metadata['atramhasis.id_generation_strategy'] = db_provider.id_generation_strategy
    metadata["id"] = db_provider.id
    return SQLAlchemyProvider(
        metadata=metadata,
        session=orm.object_session(db_provider),
        expand_strategy=db_provider.expand_strategy.value,
        uri_generator=UriPatternGenerator(db_provider.uri_pattern),
    )


@contextlib.contextmanager
def db_session(settings):
    engine = engine_from_config(settings, 'sqlalchemy.')
    session_maker = sessionmaker(bind=engine)
    session = session_maker()
    try:
        yield session
        session.commit()
    except Exception:
        session.rollback()
        raise
    finally:
        session.close()



import logging
import os

from pyramid.config import Configurator
from pyramid.config import PHASE3_CONFIG
from pyramid.interfaces import ISessionFactory
from pyramid.session import SignedCookieSessionFactory
from pyramid.settings import aslist

from atramhasis.renderers import json_renderer_verbose

LOG = logging.getLogger(__name__)


DEFAULT_SETTINGS = {
    "cache.tree.backend": "dogpile.cache.memory",
    "cache.tree.arguments.cache_size": "5000",
    "cache.tree.expiration_time": "7000",
    "cache.list.backend": "dogpile.cache.memory",
    "cache.list.arguments.cache_size": "5000",
    "cache.list.expiration_time": "7000",
    "jinja2.extensions": "jinja2.ext.do",
    "jinja2.filters": "label_sort = atramhasis.utils.label_sort",
    "dojo.mode": "dist",
    "layout.focus_conceptschemes": [],
    "skosprovider.skosregistry_factory": "atramhasis.skos.create_registry",
    "skosprovider.skosregistry_location": "request",
}


def includeme(config):
    """this function adds some configuration for the application"""
    settings = config.registry.settings
    for key, value in DEFAULT_SETTINGS.items():
        if key not in settings:
            settings[key] = value
    configure_session(config)
    config.include('pyramid_jinja2')
    config.include('pyramid_tm')
    config.add_static_view('static', 'static', cache_max_age=3600)
    config.add_renderer('csv', 'atramhasis.renderers.CSVRenderer')
    config.add_renderer('skosrenderer_verbose', json_renderer_verbose)
    # Rewrite urls with trailing slash
    config.include('pyramid_rewrite')
    config.include("pyramid_openapi3")
    config.include('atramhasis.routes')
    config.include('pyramid_skosprovider')
    config.include('atramhasis.cache')
    config.scan('pyramid_skosprovider')

    config.add_translation_dirs('atramhasis:locale/')

    config.scan()


def configure_session(config):
    """
    Configure pyramid's session factory.

    People can configure their own session factory, but if no factory is registered
    atramhasis will try configuring its own.
    """

    def check_session_factory_set():
        session_factory = config.registry.queryUtility(ISessionFactory)
        if session_factory:
            return

        settings = config.registry.settings
        if 'atramhasis.session_factory.secret' not in settings:
            msg = (
                'No session factory is configured, and '
                'atramhasis.session_factory.secret setting is missing.'
            )
            raise ValueError(msg)

        LOG.info('Using default SignedCookieSessionFactory.')
        default_session_factory = SignedCookieSessionFactory(
            settings['atramhasis.session_factory.secret']
        )
        config.registry.registerUtility(default_session_factory, ISessionFactory)

    config.action(
        'check_session_factory_set', check_session_factory_set, order=PHASE3_CONFIG + 1
    )


def main(global_config, **settings):
    """ This function returns a Pyramid WSGI application.
    """
    settings['layout.focus_conceptschemes'] = aslist(settings['layout.focus_conceptschemes'], flatten=False)

    dump_location = settings['atramhasis.dump_location']
    if not os.path.exists(dump_location):
        os.makedirs(dump_location)

    config = Configurator(settings=settings)

    return load_app(config)


def load_app(config):
    includeme(config)

    config.include('atramhasis.data:db')

    return config.make_wsgi_app()



import colander
from pyramid.view import view_config
from pyramid.view import view_defaults
from skosprovider_sqlalchemy.models import Language
from sqlalchemy.exc import NoResultFound

from atramhasis.errors import LanguageNotFoundException
from atramhasis.errors import ValidationError
from atramhasis.validators import LanguageTag
from atramhasis.validators import languagetag_validator


@view_defaults(accept='application/json', renderer='skosrenderer_verbose')
class LanguagesCrud:
    """
    This object groups CRUD REST views for languages.
    """

    def __init__(self, context, request):
        self.request = request
        self.context = context
        self.logged_in = request.authenticated_userid
        self.languages_manager = self.request.data_managers['languages_manager']

    def _get_json_body(self):
        json_body = self.request.json_body
        if 'l_id' in self.request.matchdict and 'id' not in json_body:
            json_body['id'] = self.request.matchdict['l_id']
        return json_body

    def _validate_language(self, json_language, new):
        language = LanguageTag(
            validator=languagetag_validator
        ).bind(
            request=self.request,
            new=new
        )
        try:
            return language.deserialize(json_language)
        except colander.Invalid as e:
            raise ValidationError(
                'Language could not be validated',
                e.asdict()
            )

    @view_config(route_name='atramhasis.list_languages', permission='edit')
    def list_languages(self):
        """
        Get all existing languages

        :return: list of languages
        """
        if 'sort' in self.request.params:
            sort = self.request.params['sort']
            sort_desc = (sort[0:1] == '-')
            sort = sort[1:] if sort[0:1] in ['-', '+'] else sort
            sort = sort.strip()
            languages = self.languages_manager.get_all_sorted(sort, sort_desc)
        else:
            languages = self.languages_manager.get_all()
        return languages

    @view_config(route_name='atramhasis.get_language', permission='edit')
    def get_language(self):
        """
        Get an existing language

        :raises atramhasis.errors.LanguageNotFoundException: If the language can't be found
        :return the corresponding language
        """
        l_id = self.request.matchdict['l_id']
        try:
            language = self.languages_manager.get(l_id)
        except NoResultFound:
            raise LanguageNotFoundException(l_id)

        self.request.response.status = '200'
        return language

    @view_config(route_name='atramhasis.edit_language', permission='edit')
    def edit_language(self):
        """
        Edit an existing language or create when not found

        :raises atramhasis.errors.ValidationError: If the provided json can't be validated
        """
        l_id = self.request.matchdict['l_id']
        json_body = self._get_json_body()
        json_body['id'] = l_id

        try:
            language = self.languages_manager.get(l_id)
            validated_json_language = self._validate_language(json_body, False)
            language.name = validated_json_language['name']
        except NoResultFound:
            validated_json_language = self._validate_language(json_body, True)
            language = Language(id=validated_json_language['id'], name=validated_json_language['name'])
        language = self.languages_manager.save(language)
        self.request.response.status = '200'
        return {'id': language.id, 'name': language.name}

    @view_config(route_name='atramhasis.delete_language', permission='edit')
    def delete_language(self):
        """
        Delete an existing concept

        :raises atramhasis.errors.LanguageNotFoundException: If the language can't be found
        """
        l_id = self.request.matchdict['l_id']
        try:
            language = self.languages_manager.get(l_id)
        except NoResultFound:
            raise LanguageNotFoundException(l_id)
        self.languages_manager.delete(language)

        self.request.response.status = '200'
        return {'id': language.id, 'name': language.name}



"""init

Revision ID: 3924a7ad2f8
Revises: None
Create Date: 2014-04-10 15:37:26.804103

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "3924a7ad2f8"
down_revision = None
branch_labels = ("atramhasis",)


def upgrade():
    op.create_table(
        "labeltype",
        sa.Column("name", sa.String(length=20), nullable=False),
        sa.Column("description", sa.Text(), nullable=True),
        sa.PrimaryKeyConstraint("name"),
    )
    op.create_table(
        "language",
        sa.Column("id", sa.String(length=64), nullable=False),
        sa.Column("name", sa.String(length=255), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "conceptscheme",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("uri", sa.String(length=512), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "notetype",
        sa.Column("name", sa.String(length=20), nullable=False),
        sa.Column("description", sa.Text(), nullable=True),
        sa.PrimaryKeyConstraint("name"),
    )
    op.create_table(
        "label",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("label", sa.String(length=512), nullable=False),
        sa.Column("labeltype_id", sa.String(length=20), nullable=False),
        sa.Column("language_id", sa.String(length=64), nullable=True),
        sa.ForeignKeyConstraint(
            ["labeltype_id"],
            ["labeltype.name"],
        ),
        sa.ForeignKeyConstraint(
            ["language_id"],
            ["language.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index("ix_label_labeltype_id", "label", ["labeltype_id"], unique=False)
    op.create_index("ix_label_language_id", "label", ["language_id"], unique=False)
    op.create_table(
        "note",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("note", sa.Text(), nullable=False),
        sa.Column("notetype_id", sa.String(length=20), nullable=False),
        sa.Column("language_id", sa.String(length=64), nullable=True),
        sa.ForeignKeyConstraint(
            ["language_id"],
            ["language.id"],
        ),
        sa.ForeignKeyConstraint(
            ["notetype_id"],
            ["notetype.name"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index("ix_note_language_id", "note", ["language_id"], unique=False)
    op.create_index("ix_note_notetype_id", "note", ["notetype_id"], unique=False)
    op.create_table(
        "concept",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("type", sa.String(length=30), nullable=True),
        sa.Column("concept_id", sa.Integer(), nullable=False),
        sa.Column("uri", sa.String(length=512), nullable=True),
        sa.Column("conceptscheme_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["conceptscheme_id"],
            ["conceptscheme.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index("ix_concept_concept_id", "concept", ["concept_id"], unique=False)
    op.create_index(
        "ix_concept_conceptscheme_id", "concept", ["conceptscheme_id"], unique=False
    )
    op.create_table(
        "collection_concept",
        sa.Column("collection_id", sa.Integer(), nullable=False),
        sa.Column("concept_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["collection_id"],
            ["concept.id"],
        ),
        sa.ForeignKeyConstraint(
            ["concept_id"],
            ["concept.id"],
        ),
        sa.PrimaryKeyConstraint("collection_id", "concept_id"),
    )
    op.create_table(
        "conceptscheme_note",
        sa.Column("conceptscheme_id", sa.Integer(), nullable=False),
        sa.Column("note_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["conceptscheme_id"],
            ["conceptscheme.id"],
        ),
        sa.ForeignKeyConstraint(
            ["note_id"],
            ["note.id"],
        ),
        sa.PrimaryKeyConstraint("conceptscheme_id", "note_id"),
    )
    op.create_table(
        "conceptscheme_label",
        sa.Column("conceptscheme_id", sa.Integer(), nullable=False),
        sa.Column("label_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["conceptscheme_id"],
            ["conceptscheme.id"],
        ),
        sa.ForeignKeyConstraint(
            ["label_id"],
            ["label.id"],
        ),
        sa.PrimaryKeyConstraint("conceptscheme_id", "label_id"),
    )
    op.create_table(
        "visitation",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("lft", sa.Integer(), nullable=False),
        sa.Column("rght", sa.Integer(), nullable=False),
        sa.Column("depth", sa.Integer(), nullable=False),
        sa.Column("conceptscheme_id", sa.Integer(), nullable=False),
        sa.Column("concept_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["concept_id"],
            ["concept.id"],
        ),
        sa.ForeignKeyConstraint(
            ["conceptscheme_id"],
            ["conceptscheme.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        "ix_visitation_concept_id", "visitation", ["concept_id"], unique=False
    )
    op.create_index(
        "ix_visitation_conceptscheme_id",
        "visitation",
        ["conceptscheme_id"],
        unique=False,
    )
    op.create_index("ix_visitation_depth", "visitation", ["depth"], unique=False)
    op.create_index("ix_visitation_lft", "visitation", ["lft"], unique=False)
    op.create_index("ix_visitation_rght", "visitation", ["rght"], unique=False)
    op.create_table(
        "concept_note",
        sa.Column("concept_id", sa.Integer(), nullable=False),
        sa.Column("note_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["concept_id"],
            ["concept.id"],
        ),
        sa.ForeignKeyConstraint(
            ["note_id"],
            ["note.id"],
        ),
        sa.PrimaryKeyConstraint("concept_id", "note_id"),
    )
    op.create_table(
        "concept_related_concept",
        sa.Column("concept_id_from", sa.Integer(), nullable=False),
        sa.Column("concept_id_to", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["concept_id_from"],
            ["concept.id"],
        ),
        sa.ForeignKeyConstraint(
            ["concept_id_to"],
            ["concept.id"],
        ),
        sa.PrimaryKeyConstraint("concept_id_from", "concept_id_to"),
    )
    op.create_table(
        "concept_hierarchy_concept",
        sa.Column("concept_id_broader", sa.Integer(), nullable=False),
        sa.Column("concept_id_narrower", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["concept_id_broader"],
            ["concept.id"],
        ),
        sa.ForeignKeyConstraint(
            ["concept_id_narrower"],
            ["concept.id"],
        ),
        sa.PrimaryKeyConstraint("concept_id_broader", "concept_id_narrower"),
    )
    op.create_table(
        "concept_hierarchy_collection",
        sa.Column("concept_id_broader", sa.Integer(), nullable=False),
        sa.Column("collection_id_narrower", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["collection_id_narrower"],
            ["concept.id"],
        ),
        sa.ForeignKeyConstraint(
            ["concept_id_broader"],
            ["concept.id"],
        ),
        sa.PrimaryKeyConstraint("concept_id_broader", "collection_id_narrower"),
    )
    op.create_table(
        "concept_label",
        sa.Column("concept_id", sa.Integer(), nullable=False),
        sa.Column("label_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["concept_id"],
            ["concept.id"],
        ),
        sa.ForeignKeyConstraint(
            ["label_id"],
            ["label.id"],
        ),
        sa.PrimaryKeyConstraint("concept_id", "label_id"),
    )


def downgrade():
    op.drop_table("concept_label")
    op.drop_table("concept_hierarchy_collection")
    op.drop_table("concept_hierarchy_concept")
    op.drop_table("concept_related_concept")
    op.drop_table("concept_note")
    op.drop_index("ix_visitation_rght", table_name="visitation")
    op.drop_index("ix_visitation_lft", table_name="visitation")
    op.drop_index("ix_visitation_depth", table_name="visitation")
    op.drop_index("ix_visitation_conceptscheme_id", table_name="visitation")
    op.drop_index("ix_visitation_concept_id", table_name="visitation")
    op.drop_table("visitation")
    op.drop_table("conceptscheme_label")
    op.drop_table("conceptscheme_note")
    op.drop_table("collection_concept")
    op.drop_index("ix_concept_conceptscheme_id", table_name="concept")
    op.drop_index("ix_concept_concept_id", table_name="concept")
    op.drop_table("concept")
    op.drop_index("ix_note_notetype_id", table_name="note")
    op.drop_index("ix_note_language_id", table_name="note")
    op.drop_table("note")
    op.drop_index("ix_label_language_id", table_name="label")
    op.drop_index("ix_label_labeltype_id", table_name="label")
    op.drop_table("label")
    op.drop_table("notetype")
    op.drop_table("conceptscheme")
    op.drop_table("language")
    op.drop_table("labeltype")



import json
import os
import sys
from datetime import date
from datetime import datetime

from pyramid.paster import get_appsettings
from pyramid.paster import setup_logging
from pyramid.scripts.common import parse_vars
from skosprovider.providers import VocabularyProvider
from skosprovider_sqlalchemy import utils as skosprovider_utils
from skosprovider_sqlalchemy.models import ConceptScheme
from skosprovider_sqlalchemy.models import Label
from sqlalchemy import engine_from_config
from sqlalchemy.orm import Session

from atramhasis.data.models import ExpandStrategy
from atramhasis.data.models import IDGenerationStrategy
from atramhasis.data.models import Provider


def usage(argv):
    cmd = os.path.basename(argv[0])
    print('usage: %s <config_uri> [var=value]\n'
          '(example: "%s development.ini")' % (cmd, cmd))
    sys.exit(1)


def json_serial(obj):
    if isinstance(obj, (datetime, date)):
        return obj.isoformat()
    raise TypeError(f"Type {obj} not serializable")


def initialize_providers(session):
    from fixtures.data import trees, geo
    from fixtures.styles_and_cultures import styles_and_cultures
    from fixtures.materials import materials
    from fixtures.eventtypes import eventtypes
    from fixtures.heritagetypes import heritagetypes
    from fixtures.periods import periods
    from fixtures.species import species
    from fixtures.bluebirds import bluebirds

    import_provider(
        trees,
        session,
        ConceptScheme(
            id=1,
            uri='urn:x-skosprovider:trees',
            labels=[
                Label('Verschillende soorten bomen', 'prefLabel', 'nl'),
                Label('Different types of trees', 'prefLabel', 'en')
            ]
        ),
    )
    import_provider(
        geo,
        session,
        ConceptScheme(
            id=2,
            uri='urn:x-skosprovider:geo',
            labels=[
                Label('Geografie', 'prefLabel', 'nl'),
                Label('Geography', 'prefLabel', 'en')
            ]
        ),
    )
    import_provider(
        styles_and_cultures,
        session,
        ConceptScheme(
            id=3,
            uri='https://id.erfgoed.net/thesauri/stijlen_en_culturen',
            labels=[
                Label('Stijlen en Culturen', 'prefLabel', 'nl'),
                Label('Styles and Cultures', 'prefLabel', 'en')
            ]
        ),
    )
    import_provider(
        materials,
        session,
        ConceptScheme(
            id=4,
            uri='https://id.erfgoed.net/thesauri/materialen',
            labels=[
                Label('Materialen', 'prefLabel', 'nl'),
                Label('Materials', 'prefLabel', 'en')
            ]
        ),
    )
    import_provider(
        eventtypes,
        session,
        ConceptScheme(
            id=5,
            uri='https://id.erfgoed.net/thesauri/gebeurtenistypes',
            labels=[
                Label('Gebeurtenistypes', 'prefLabel', 'nl'),
                Label('Event types', 'prefLabel', 'en')
            ]
        ),
    )
    import_provider(
        heritagetypes,
        session,
        ConceptScheme(
            id=6,
            uri='https://id.erfgoed.net/thesauri/erfgoedtypes',
            labels=[
                Label('Erfgoedtypes', 'prefLabel', 'nl'),
                Label('Heritage types', 'prefLabel', 'en')
            ]
        ),
    )
    import_provider(
        periods,
        session,
        ConceptScheme(
            id=7,
            uri='https://id.erfgoed.net/thesauri/dateringen',
            labels=[
                Label('Dateringen', 'prefLabel', 'nl'),
                Label('Periods', 'prefLabel', 'en')
            ]
        ),
    )
    import_provider(
        species,
        session,
        ConceptScheme(
            id=8,
            uri='https://id.erfgoed.net/thesauri/soorten',
            labels=[
                Label('Soorten', 'prefLabel', 'nl'),
                Label('Species', 'prefLabel', 'en')
            ]
        ),
    )
    import_provider(
        bluebirds,
        session,
        ConceptScheme(
            id=9,
            uri='https://id.bluebirds.org',
            labels=[
                Label('Blauwe vogels', 'prefLabel', 'nl'),
                Label('Blue birds', 'prefLabel', 'en')
            ]
        ),
    )


def import_provider(
    provider: VocabularyProvider, session: Session, conceptscheme: ConceptScheme
):
    concept_scheme = skosprovider_utils.import_provider(
        provider, session, conceptscheme=conceptscheme,
    )

    if provider.uri_generator:
        uri_pattern = getattr(provider.uri_generator, 'pattern', None)
    else:
        uri_pattern = None

    db_provider = Provider()
    db_provider.meta = json.loads(json.dumps(provider.metadata, default=json_serial))
    db_provider.id_generation_strategy = IDGenerationStrategy.NUMERIC
    db_provider.expand_strategy = ExpandStrategy.RECURSE
    db_provider.conceptscheme = concept_scheme
    db_provider.id = provider.get_vocabulary_id()
    db_provider.uri_pattern = uri_pattern
    if 'conceptscheme_id' in db_provider.meta:
        del db_provider.meta['conceptscheme_id']

    session.add(db_provider)


def main(argv=sys.argv):
    if len(argv) < 2:
        usage(argv)

    config_uri = argv[1]
    options = parse_vars(argv[2:])
    setup_logging(config_uri)
    settings = get_appsettings(config_uri, options=options)

    engine = engine_from_config(settings, 'sqlalchemy.')
    with Session(engine) as session:
        initialize_providers(session)
        session.commit()
    print('--atramhasis-db-initialized--')


if __name__ == '__main__':
    main()


from typing import Mapping

from skosprovider.registry import Registry
from sqlalchemy.exc import NoResultFound
from sqlalchemy.orm import Session

from atramhasis import mappers
from atramhasis.data.datamanagers import ProviderDataManager
from atramhasis.data.models import Provider
from atramhasis.errors import SQLAlchemyProviderNotFoundException
from atramhasis.errors import ValidationError
from atramhasis.scripts.delete_scheme import delete_scheme
from atramhasis.validators import validate_provider_json


def create_provider(json_data: Mapping, session: Session, skos_registry: Registry) -> Provider:
    """Process a provider JSON into a newly stored Provider."""
    for provider in skos_registry.get_providers():
        if provider.get_vocabulary_uri() == json_data["conceptscheme_uri"]:
            raise ValidationError(
                "Provider could not be validated.",
                [{"conceptscheme_uri": "Collides with existing provider."}],
            )
    validate_provider_json(json_data)
    db_provider = mappers.map_provider(json_data)
    if not db_provider.id:
        # Store conceptscheme first so we can copy its id
        session.add(db_provider.conceptscheme)
        session.flush()
        db_provider.id = str(db_provider.conceptscheme.id)

    session.add(db_provider)
    session.flush()

    return db_provider


def update_provider(provider_id: str, json_data: Mapping, session: Session) -> Provider:
    """Process a JSON into to update an existing provider."""
    validate_provider_json(json_data, provider_id)
    manager = ProviderDataManager(session)
    db_provider = manager.get_provider_by_id(provider_id)
    db_provider = mappers.map_provider(json_data, provider=db_provider)
    session.flush()
    return db_provider


def delete_provider(provider_id, session: Session) -> None:
    manager = ProviderDataManager(session)
    try:
        db_provider = manager.get_provider_by_id(provider_id)
    except NoResultFound:
        raise SQLAlchemyProviderNotFoundException(provider_id)
    delete_scheme(session, db_provider.conceptscheme.id)
    session.delete(db_provider)



"""cascades

Revision ID: 3e9675b35dfc
Revises: cb568ec81000
Create Date: 2020-08-12 14:48:35.592316

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "3e9675b35dfc"
down_revision = "cb568ec81000"


def _get_convention():
    is_sqlite = op.get_bind().dialect.name == "sqlite"
    return (
        {"fk": "FK_%(table_name)s_%(referred_table_name)s"}
        if is_sqlite
        else {"fk": "%(table_name)s_%(column_0_name)s_fkey"}
    )


def upgrade():

    args = [
        ("conceptscheme_source", "conceptscheme", "conceptscheme_id", "id"),
        ("conceptscheme_source", "source", "source_id", "id"),
        ("concept_source", "concept", "concept_id", "id"),
        ("concept_source", "source", "source_id", "id"),
        ("conceptscheme_language", "conceptscheme", "conceptscheme_id", "id"),
        ("match", "concept", "concept_id", "id"),
        ("label", "language", "language_id", "id"),
        ("note", "language", "language_id", "id"),
        ("collection_concept", "concept", "concept_id", "id"),
        ("collection_concept", "concept", "collection_id", "id"),
        ("conceptscheme_note", "conceptscheme", "conceptscheme_id", "id"),
        ("conceptscheme_label", "conceptscheme", "conceptscheme_id", "id"),
        ("visitation", "concept", "concept_id", "id"),
        ("visitation", "conceptscheme", "conceptscheme_id", "id"),
        ("concept_note", "concept", "concept_id", "id"),
        ("concept_note", "note", "note_id", "id"),
        ("concept_related_concept", "concept", "concept_id_from", "id"),
        ("concept_related_concept", "concept", "concept_id_to", "id"),
        ("concept_hierarchy_concept", "concept", "concept_id_narrower", "id"),
        ("concept_hierarchy_concept", "concept", "concept_id_broader", "id"),
        ("concept_hierarchy_collection", "concept", "concept_id_broader", "id"),
        ("concept_hierarchy_collection", "concept", "collection_id_narrower", "id"),
        ("concept_label", "concept", "concept_id", "id"),
        ("concept_label", "label", "label_id", "id"),
    ]
    convention = _get_convention()

    for source_table, referent_table, local_col, remote_col in args:
        with op.batch_alter_table(
            source_table, naming_convention=convention
        ) as batch_op:
            constraint_name = convention["fk"] % {
                "table_name": source_table,
                "referred_table_name": referent_table,
                "column_0_name": local_col,
            }
            batch_op.drop_constraint(constraint_name)
            batch_op.create_foreign_key(
                constraint_name,
                referent_table,
                [local_col],
                [remote_col],
                ondelete="cascade",
            )


def downgrade():
    pass



import unittest
from unittest.mock import MagicMock
from unittest.mock import Mock

from pyramid import testing
from skosprovider.providers import VocabularyProvider
from skosprovider.registry import Registry
from skosprovider_getty.providers import AATProvider
from skosprovider_sqlalchemy.models import Collection
from skosprovider_sqlalchemy.models import Concept
from skosprovider_sqlalchemy.models import ConceptScheme
from skosprovider_sqlalchemy.models import Label
from skosprovider_sqlalchemy.models import Language
from skosprovider_sqlalchemy.models import Match
from skosprovider_sqlalchemy.models import MatchType
from skosprovider_sqlalchemy.models import Note
from skosprovider_sqlalchemy.models import Source
from skosprovider_sqlalchemy.providers import SQLAlchemyProvider

from atramhasis import renderers
from fixtures.data import trees


class TestJsonRenderer(unittest.TestCase):

    def setUp(self):
        self.concept = Concept()
        self.concept.id = 11
        self.concept.concept_id = 101
        self.concept.uri = 'urn:x-atramhasis-demo:TREES:101'
        self.concept.conceptscheme_id = 1

        notes = []
        note = Note(note='test note', notetype_id='example', language_id='en')
        note2 = Note(note='note def', notetype_id='definition', language_id='en')
        notes.append(note)
        notes.append(note2)
        self.concept.notes = notes

        labels = []
        label = Label(label='een label', labeltype_id='prefLabel', language_id='nl')
        label2 = Label(label='other label', labeltype_id='altLabel', language_id='en')
        label3 = Label(label='and some other label', labeltype_id='altLabel', language_id='en')
        labels.append(label)
        labels.append(label2)
        labels.append(label3)
        self.concept.labels = labels

        sources = []
        source = Source('Van Daele K. 2009')
        sources.append(source)
        self.concept.sources = sources

        matches = []
        match = Match()
        match.matchtype = MatchType(name='closeMatch', description='test')
        match.uri = 'urn:somethingelse:st1'
        matches.append(match)
        match2 = Match()
        match2.matchtype = MatchType(name='closeMatch', description='test')
        match2.uri = 'urn:somethingelse:st2'
        matches.append(match2)
        match3 = Match()
        match3.matchtype = MatchType(name='exactMatch', description='test')
        match3.uri = 'urn:something:thingy'
        matches.append(match3)
        self.concept.matches = matches

        self.collection = Collection()
        self.collection.id = 12
        self.collection.concept_id = 102
        self.collection.uri = 'urn:x-atramhasis-demo:TREES:102'
        self.collection.conceptscheme_id = 1

        self.conceptscheme = ConceptScheme()
        self.conceptscheme.id = 1
        self.conceptscheme.labels = labels
        self.conceptscheme.notes = notes
        self.conceptscheme.sources = sources
        self.conceptscheme.uri = None

        self.regis = Registry()
        self.regis.register_provider(trees)
        self.request = testing.DummyRequest()
        self.request.skos_registry = self.regis
        self.request.matchdict = {'scheme_id': 'TREES'}
        self.request.locale_name = 'nl'

        self.concept.member_of.add(self.collection)
        self.collection.members.add(self.concept)

    def test_label_adapter(self):
        from atramhasis.renderers import label_adapter
        label = self.concept.labels[2]
        label = label_adapter(label, {})
        self.assertIsInstance(label, dict)
        self.assertEqual(label['label'], 'and some other label')
        self.assertEqual(label['type'], 'altLabel')
        self.assertEqual(label['language'], 'en')

    def test_note_adapter(self):
        from atramhasis.renderers import note_adapter
        n = self.concept.notes[0]
        note = note_adapter(n, {})
        self.assertIsInstance(note, dict)
        self.assertEqual(note['note'], 'test note')
        self.assertEqual(note['type'], 'example')
        self.assertEqual(note['language'], 'en')

    def test_source_adapter(self):
        from atramhasis.renderers import source_adapter
        s = self.concept.sources[0]
        source = source_adapter(s, {})
        self.assertIsInstance(source, dict)
        self.assertEqual(source['citation'], 'Van Daele K. 2009')

    def test_concept_adapter(self):
        from atramhasis.renderers import concept_adapter
        c = self.concept
        concept = concept_adapter(c, self.request)
        self.assertIsInstance(concept, dict)
        self.assertEqual(concept['id'], 101)
        self.assertEqual(concept['type'], 'concept')
        self.assertEqual(concept['uri'], 'urn:x-atramhasis-demo:TREES:101')
        self.assertIsNotNone(concept['label'], 'een label')
        self.assertIsInstance(concept['labels'], list)
        self.assertEqual(len(concept['labels']), 3)
        self.assertIsInstance(concept['notes'], list)
        self.assertEqual(len(concept['notes']), 2)
        self.assertIsInstance(concept['sources'], list)
        self.assertEqual(len(concept['sources']), 1)
        self.assertIsInstance(concept['member_of'], list)
        self.assertEqual(len(concept['member_of']), 1)
        self.assertIsInstance(concept['narrower'], list)
        self.assertEqual(len(concept['narrower']), 0)
        self.assertIsInstance(concept['broader'], list)
        self.assertEqual(len(concept['broader']), 0)
        self.assertIsInstance(concept['related'], list)
        self.assertEqual(len(concept['related']), 0)
        self.assertIsInstance(concept['matches'], dict)
        self.assertEqual(len(concept['matches']['exact']), 1)
        self.assertEqual(len(concept['matches']['close']), 2)

    def test_collection_adapter(self):
        from atramhasis.renderers import collection_adapter
        c = self.collection
        collection = collection_adapter(c, self.request)
        self.assertIsInstance(collection, dict)
        self.assertEqual(collection['id'], 102)
        self.assertEqual(collection['type'], 'collection')
        self.assertEqual(collection['uri'], 'urn:x-atramhasis-demo:TREES:102')
        self.assertIsNone(collection['label'])
        self.assertIsInstance(collection['labels'], list)
        self.assertEqual(len(collection['labels']), 0)
        self.assertIsInstance(collection['member_of'], list)
        self.assertEqual(len(collection['member_of']), 0)
        self.assertIsInstance(collection['members'], list)
        self.assertEqual(len(collection['members']), 1)

    def test_map_relation_concept(self):
        from atramhasis.renderers import map_relation
        c = self.concept
        relation = map_relation(c)
        self.assertIsInstance(relation, dict)
        self.assertEqual(relation['id'], 101)
        self.assertEqual(relation['type'], 'concept')
        self.assertEqual(relation['uri'], 'urn:x-atramhasis-demo:TREES:101')
        self.assertIsNotNone(relation['label'], 'een label')
        self.assertRaises(KeyError, lambda: relation['notes'])
        self.assertRaises(KeyError, lambda: relation['member_of'])
        self.assertRaises(KeyError, lambda: relation['narrower'])
        self.assertRaises(KeyError, lambda: relation['broader'])
        self.assertRaises(KeyError, lambda: relation['related'])

    def test_map_relation_collection(self):
        from atramhasis.renderers import map_relation
        c = self.collection
        relation = map_relation(c)
        self.assertIsInstance(relation, dict)
        self.assertEqual(relation['id'], 102)
        self.assertEqual(relation['type'], 'collection')
        self.assertEqual(relation['uri'], 'urn:x-atramhasis-demo:TREES:102')
        self.assertIsNone(relation['label'])
        self.assertRaises(KeyError, lambda: relation['members'])
        self.assertRaises(KeyError, lambda: relation['member_of'])

    def test_language_adaptor(self):
        from atramhasis.renderers import language_adaptor
        language = Language(id='af', name='Afrikaans')
        res = language_adaptor(language, None)
        self.assertIsNotNone(res)
        self.assertIsInstance(res, dict)
        self.assertEqual(res['id'], 'af')
        self.assertEqual(res['name'], 'Afrikaans')

    def test_conceptscheme_adapter(self):
        from atramhasis.renderers import conceptscheme_adapter
        c = self.conceptscheme
        conceptscheme = conceptscheme_adapter(c, self.request)
        self.assertGreater(len(conceptscheme['notes']), 0)
        self.assertGreater(len(conceptscheme['labels']), 0)
        self.assertGreater(len(conceptscheme['sources']), 0)
        self.assertIsNone(conceptscheme['uri'])
        self.assertEqual('een label', conceptscheme['label'])
        self.assertEqual(1, conceptscheme['id'])
        self.assertEqual(0, len(conceptscheme['subject']))
        self.assertIsInstance(conceptscheme, dict)

    def test_conceptscheme_language_handling(self):
        from atramhasis.renderers import conceptscheme_adapter
        c = self.conceptscheme
        conceptscheme = conceptscheme_adapter(c, self.request)
        self.assertEqual('een label', conceptscheme['label'])
        self.request.locale_name = 'en'
        conceptscheme = conceptscheme_adapter(c, self.request)
        self.assertIn(conceptscheme['label'], ['other label', 'and some other label'])

    def test_provider_adapter(self):
        provider = VocabularyProvider(
            metadata={
                'id': 'provider-id',
                'default_language': 'nl-be',
                'subject': 'sub',
                'atramhasis.force_display_language': 'force-nl'
            }
        )
        result = renderers.provider_adapter(provider)
        self.assertEqual(
            {
                'conceptscheme_uri': 'urn:x-skosprovider:provider-id',
                'default_language': 'nl-be',
                'force_display_language': 'force-nl',
                'id': 'provider-id',
                'subject': 'sub',
                'type': 'VocabularyProvider',
                'uri_pattern': 'urn:x-skosprovider:%s:%s',
                'metadata': {},
            },
            result
        )

    def test_provider_adapter_aat_provider(self):
        provider = AATProvider(
            metadata={
                'id': 'provider-id',
                'default_language': 'nl-be',
                'subject': 'sub',
                'atramhasis.force_display_language': 'force-nl'
            }
        )
        result = renderers.provider_adapter(provider)
        self.assertEqual(
            {
                'conceptscheme_uri': 'http://vocab.getty.edu/aat/',
                'default_language': 'nl-be',
                'force_display_language': 'force-nl',
                'id': 'provider-id',
                'subject': 'sub',
                'type': 'AATProvider',
                'uri_pattern': None,
                'metadata': {'uri': 'http://vocab.getty.edu/aat/'},
            },
            result
        )

    def test_provider_adapter_sqlalchemy_provider(self):
        sessionmaker = Mock()
        session = sessionmaker()
        session.get.return_value = ConceptScheme(
            id=1,
            uri='urn:x-skosprovider:trees',
            labels=[
                Label('Verschillende soorten bomen', 'prefLabel', 'nl'),
                Label('Different types of trees', 'prefLabel', 'en')
            ]
        )
        provider = SQLAlchemyProvider(
            metadata={
                'id': 'provider-id',
                'default_language': 'nl-be',
                'subject': 'sub',
                'atramhasis.force_display_language': 'force-nl',
                'conceptscheme_id': 1,
            },
            session=sessionmaker,
        )
        result = renderers.sa_provider_adapter(provider)
        self.assertEqual(
            {
                'conceptscheme_uri': 'urn:x-skosprovider:trees',
                'default_language': 'nl-be',
                'force_display_language': 'force-nl',
                'id': 'provider-id',
                'expand_strategy': 'recurse',
                'id_generation_strategy': 'NUMERIC',
                'subject': 'sub',
                'type': 'SQLAlchemyProvider',
                'uri_pattern': 'urn:x-skosprovider:%s:%s',
                'metadata': {},
            },
            result
        )



import subprocess
from pathlib import Path

from hatchling.builders.hooks.plugin.interface import BuildHookInterface


class BuildHook(BuildHookInterface):
    def initialize(self, version: str, build_data) -> None:
        """Build frontend when building the wheel."""
        super().initialize(version, build_data)

        root_dir = Path(__file__).parent
        static = root_dir / "atramhasis" / "static"
        static_admin = static / "admin"
        subprocess.run(["npm", "install"], cwd=static, check=True)
        subprocess.run(["npm", "install"], cwd=static_admin, check=True)
        subprocess.run(["grunt", "-v", "build"], cwd=static_admin, check=True)



import logging
import unittest
from unittest.mock import MagicMock
from unittest.mock import Mock

from pyramid.response import Response
from testfixtures import LogCapture

from atramhasis.audit import _origin_from_response
from atramhasis.audit import audit

log = logging.getLogger('')


cs_mock = MagicMock()
cs_mock.get_metadata = Mock(return_value={'subject': ['external']})


class RecordingManager:
    def __init__(self):
        self.saved_objects = []

    def save(self, db_object):
        self.saved_objects.append(db_object)


class DummyParent:
    def __init__(self):
        self.request = MagicMock()

    @audit
    def dummy(self):
        return None

    @audit
    def dummy_with_response(self):
        return Response(content_type='application/rdf+xml')


class AuditTests(unittest.TestCase):
    def setUp(self):
        self.audit_manager = RecordingManager()
        self.dummy_parent = DummyParent()
        self.dummy_parent.request.data_managers = {
            'audit_manager': self.audit_manager
        }

    def tearDown(self):
        pass

    def _check(self, nr, origin, type_id_list):
        self.assertEqual(nr, len(self.audit_manager.saved_objects))
        self.assertEqual(origin, self.audit_manager.saved_objects[nr - 1].origin)
        for type_id in type_id_list:
            self.assertEqual('1', getattr(self.audit_manager.saved_objects[nr - 1], type_id))

    def test_audit_rest(self):
        self.dummy_parent.request.url = "http://host/conceptschemes/STYLES"
        self.dummy_parent.request.accept = ['application/json']
        self.dummy_parent.request.matchdict = {'scheme_id': '1'}
        self.dummy_parent.dummy()
        self._check(1, 'REST', ['conceptscheme_id'])
        self.dummy_parent.request.matchdict = {'scheme_id': '1', 'c_id': '1'}
        self.dummy_parent.dummy()
        self._check(2, 'REST', ['conceptscheme_id', 'concept_id'])

    def test_audit_html(self):
        self.dummy_parent.request.url = "http://host/conceptschemes/STYLES"
        self.dummy_parent.request.accept = ['text/html']
        self.dummy_parent.request.matchdict = {'scheme_id': '1'}
        self.dummy_parent.dummy()
        self._check(1, 'HTML', ['conceptscheme_id'])
        self.dummy_parent.request.matchdict = {'scheme_id': '1', 'c_id': '1'}
        self.dummy_parent.dummy()
        self._check(2, 'HTML', ['conceptscheme_id', 'concept_id'])

    def test_audit_rdf_xml(self):
        self.dummy_parent.request.matchdict = {'scheme_id': '1'}
        self.dummy_parent.dummy_with_response()
        self._check(1, 'RDF', ['conceptscheme_id'])
        self.dummy_parent.request.matchdict = {'scheme_id': '1', 'c_id': '1'}
        self.dummy_parent.dummy_with_response()
        self._check(2, 'RDF', ['conceptscheme_id', 'concept_id'])

    def test_audit_csv(self):
        self.dummy_parent.request.url = "http://host/conceptschemes/STYLES.csv"
        self.dummy_parent.request.accept = "text/csv"
        self.dummy_parent.request.matchdict = {'scheme_id': '1'}
        self.dummy_parent.dummy()
        self._check(1, 'CSV', ['conceptscheme_id'])
        self.dummy_parent.request.matchdict = {'scheme_id': '1', 'c_id': '1'}
        self.dummy_parent.dummy()
        self._check(2, 'CSV', ['conceptscheme_id', 'concept_id'])

    def test_audit_other(self):
        self.dummy_parent.request.url = "http://host/conceptschemes/STYLES"
        self.dummy_parent.request.accept = ['application/octet-stream']
        self.dummy_parent.request.matchdict = {'scheme_id': '1'}
        self.dummy_parent.dummy()
        self._check(1, None, ['conceptscheme_id'])
        self.dummy_parent.request.matchdict = {'scheme_id': '1', 'c_id': '1'}
        self.dummy_parent.dummy()
        self._check(2, None, ['conceptscheme_id', 'concept_id'])

    def test_audit_do_not_record_external(self):
        self.dummy_parent.request.url = "http://host/conceptschemes/EXT"
        self.dummy_parent.request.accept = ['application/octet-stream']
        self.dummy_parent.request.matchdict = {'scheme_id': '1'}
        self.dummy_parent.request.skos_registry.get_provider = Mock(return_value=cs_mock)
        self.dummy_parent.dummy()
        self.assertEqual(0, len(self.audit_manager.saved_objects))
        self.dummy_parent.request.matchdict = {'scheme_id': '1', 'c_id': '1'}
        self.dummy_parent.dummy()
        self.assertEqual(0, len(self.audit_manager.saved_objects))

    def test_invalid_use(self):
        with LogCapture() as logs:
            self.dummy_parent.request.url = "http://host/conceptschemes/STYLES"
            self.dummy_parent.request.accept = ['application/json']
            self.dummy_parent.request.matchdict = {'invalid_parameter_id': '1'}
            self.dummy_parent.dummy()
        self.assertIn('Misuse of the audit decorator. The url must at least contain a {scheme_id} parameter', str(logs))

    def test_origin_from_response_None(self):
        res = Response(content_type='application/octet-stream')
        self.assertIsNone(_origin_from_response(res))
